# Slixmpp: The Slick XMPP Library
# Copyright © 2021 Mathieu Pasquet <mathieui@mathieui.net>
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.

from typing import TypedDict

# Plugins mega-dict

from slixmpp.plugins.xep_0004 import XEP_0004
from slixmpp.plugins.xep_0009 import XEP_0009
from slixmpp.plugins.xep_0012 import XEP_0012
from slixmpp.plugins.xep_0013 import XEP_0013
from slixmpp.plugins.xep_0020 import XEP_0020
from slixmpp.plugins.xep_0027 import XEP_0027
from slixmpp.plugins.xep_0030 import XEP_0030
from slixmpp.plugins.xep_0033 import XEP_0033
from slixmpp.plugins.xep_0045 import XEP_0045
from slixmpp.plugins.xep_0047 import XEP_0047
from slixmpp.plugins.xep_0048 import XEP_0048
from slixmpp.plugins.xep_0049 import XEP_0049
from slixmpp.plugins.xep_0050 import XEP_0050
from slixmpp.plugins.xep_0054 import XEP_0054
from slixmpp.plugins.xep_0059 import XEP_0059
from slixmpp.plugins.xep_0060 import XEP_0060
from slixmpp.plugins.xep_0065 import XEP_0065
from slixmpp.plugins.xep_0066 import XEP_0066
from slixmpp.plugins.xep_0070 import XEP_0070
from slixmpp.plugins.xep_0071 import XEP_0071
from slixmpp.plugins.xep_0077 import XEP_0077
from slixmpp.plugins.xep_0079 import XEP_0079
from slixmpp.plugins.xep_0080 import XEP_0080
from slixmpp.plugins.xep_0082 import XEP_0082
from slixmpp.plugins.xep_0084 import XEP_0084
from slixmpp.plugins.xep_0085 import XEP_0085
from slixmpp.plugins.xep_0086 import XEP_0086
from slixmpp.plugins.xep_0092 import XEP_0092
from slixmpp.plugins.xep_0106 import XEP_0106
from slixmpp.plugins.xep_0107 import XEP_0107
from slixmpp.plugins.xep_0108 import XEP_0108
from slixmpp.plugins.xep_0115 import XEP_0115
from slixmpp.plugins.xep_0118 import XEP_0118
from slixmpp.plugins.xep_0122 import XEP_0122
from slixmpp.plugins.xep_0128 import XEP_0128
from slixmpp.plugins.xep_0131 import XEP_0131
from slixmpp.plugins.xep_0133 import XEP_0133
from slixmpp.plugins.xep_0152 import XEP_0152
from slixmpp.plugins.xep_0153 import XEP_0153
from slixmpp.plugins.xep_0163 import XEP_0163
from slixmpp.plugins.xep_0172 import XEP_0172
from slixmpp.plugins.xep_0184 import XEP_0184
from slixmpp.plugins.xep_0186 import XEP_0186
from slixmpp.plugins.xep_0191 import XEP_0191
from slixmpp.plugins.xep_0196 import XEP_0196
from slixmpp.plugins.xep_0198 import XEP_0198
from slixmpp.plugins.xep_0199 import XEP_0199
from slixmpp.plugins.xep_0202 import XEP_0202
from slixmpp.plugins.xep_0203 import XEP_0203
from slixmpp.plugins.xep_0221 import XEP_0221
from slixmpp.plugins.xep_0222 import XEP_0222
from slixmpp.plugins.xep_0223 import XEP_0223
from slixmpp.plugins.xep_0224 import XEP_0224
from slixmpp.plugins.xep_0231 import XEP_0231
from slixmpp.plugins.xep_0235 import XEP_0235
from slixmpp.plugins.xep_0249 import XEP_0249
from slixmpp.plugins.xep_0256 import XEP_0256
from slixmpp.plugins.xep_0257 import XEP_0257
from slixmpp.plugins.xep_0258 import XEP_0258
from slixmpp.plugins.xep_0264 import XEP_0264
from slixmpp.plugins.xep_0279 import XEP_0279
from slixmpp.plugins.xep_0280 import XEP_0280
from slixmpp.plugins.xep_0297 import XEP_0297
from slixmpp.plugins.xep_0300 import XEP_0300
from slixmpp.plugins.xep_0308 import XEP_0308
from slixmpp.plugins.xep_0313 import XEP_0313
from slixmpp.plugins.xep_0317 import XEP_0317
from slixmpp.plugins.xep_0319 import XEP_0319
from slixmpp.plugins.xep_0332 import XEP_0332
from slixmpp.plugins.xep_0333 import XEP_0333
from slixmpp.plugins.xep_0334 import XEP_0334
from slixmpp.plugins.xep_0335 import XEP_0335
from slixmpp.plugins.xep_0352 import XEP_0352
from slixmpp.plugins.xep_0353 import XEP_0353
from slixmpp.plugins.xep_0359 import XEP_0359
from slixmpp.plugins.xep_0363 import XEP_0363
from slixmpp.plugins.xep_0369 import XEP_0369
from slixmpp.plugins.xep_0372 import XEP_0372
from slixmpp.plugins.xep_0377 import XEP_0377
from slixmpp.plugins.xep_0380 import XEP_0380
from slixmpp.plugins.xep_0382 import XEP_0382
from slixmpp.plugins.xep_0385 import XEP_0385
from slixmpp.plugins.xep_0394 import XEP_0394
from slixmpp.plugins.xep_0402 import XEP_0402
from slixmpp.plugins.xep_0403 import XEP_0403
from slixmpp.plugins.xep_0404 import XEP_0404
from slixmpp.plugins.xep_0405 import XEP_0405
from slixmpp.plugins.xep_0410 import XEP_0410
from slixmpp.plugins.xep_0421 import XEP_0421
from slixmpp.plugins.xep_0422 import XEP_0422
from slixmpp.plugins.xep_0424 import XEP_0424
from slixmpp.plugins.xep_0425 import XEP_0425
from slixmpp.plugins.xep_0428 import XEP_0428
from slixmpp.plugins.xep_0437 import XEP_0437
from slixmpp.plugins.xep_0439 import XEP_0439
from slixmpp.plugins.xep_0441 import XEP_0441
from slixmpp.plugins.xep_0444 import XEP_0444
from slixmpp.plugins.xep_0446 import XEP_0446
from slixmpp.plugins.xep_0447 import XEP_0447
from slixmpp.plugins.xep_0449 import XEP_0449
from slixmpp.plugins.xep_0454 import XEP_0454
from slixmpp.plugins.xep_0455 import XEP_0455
from slixmpp.plugins.xep_0461 import XEP_0461
from slixmpp.plugins.xep_0463 import XEP_0463
from slixmpp.plugins.xep_0469 import XEP_0469
from slixmpp.plugins.xep_0482 import XEP_0482
from slixmpp.plugins.xep_0490 import XEP_0490
from slixmpp.plugins.xep_0492 import XEP_0492
from slixmpp.plugins.xep_0494 import XEP_0494
from slixmpp.plugins.xep_0502 import XEP_0502
from slixmpp.plugins.xep_0511 import XEP_0511
from slixmpp.plugins.xep_0513 import XEP_0513


class PluginsDict(TypedDict):
    xep_0004: XEP_0004
    xep_0009: XEP_0009
    xep_0012: XEP_0012
    xep_0013: XEP_0013
    xep_0020: XEP_0020
    xep_0027: XEP_0027
    xep_0030: XEP_0030
    xep_0033: XEP_0033
    xep_0045: XEP_0045
    xep_0047: XEP_0047
    xep_0048: XEP_0048
    xep_0049: XEP_0049
    xep_0050: XEP_0050
    xep_0054: XEP_0054
    xep_0059: XEP_0059
    xep_0060: XEP_0060
    xep_0065: XEP_0065
    xep_0066: XEP_0066
    xep_0070: XEP_0070
    xep_0071: XEP_0071
    xep_0077: XEP_0077
    xep_0079: XEP_0079
    xep_0080: XEP_0080
    xep_0082: XEP_0082
    xep_0084: XEP_0084
    xep_0085: XEP_0085
    xep_0086: XEP_0086
    xep_0092: XEP_0092
    xep_0106: XEP_0106
    xep_0107: XEP_0107
    xep_0108: XEP_0108
    xep_0115: XEP_0115
    xep_0118: XEP_0118
    xep_0122: XEP_0122
    xep_0128: XEP_0128
    xep_0131: XEP_0131
    xep_0133: XEP_0133
    xep_0152: XEP_0152
    xep_0153: XEP_0153
    xep_0163: XEP_0163
    xep_0172: XEP_0172
    xep_0184: XEP_0184
    xep_0186: XEP_0186
    xep_0191: XEP_0191
    xep_0196: XEP_0196
    xep_0198: XEP_0198
    xep_0199: XEP_0199
    xep_0202: XEP_0202
    xep_0203: XEP_0203
    xep_0221: XEP_0221
    xep_0222: XEP_0222
    xep_0223: XEP_0223
    xep_0224: XEP_0224
    xep_0231: XEP_0231
    xep_0235: XEP_0235
    xep_0249: XEP_0249
    xep_0256: XEP_0256
    xep_0257: XEP_0257
    xep_0258: XEP_0258
    xep_0264: XEP_0264
    xep_0279: XEP_0279
    xep_0280: XEP_0280
    xep_0297: XEP_0297
    xep_0300: XEP_0300
    xep_0308: XEP_0308
    xep_0313: XEP_0313
    xep_0317: XEP_0317
    xep_0319: XEP_0319
    xep_0332: XEP_0332
    xep_0333: XEP_0333
    xep_0334: XEP_0334
    xep_0335: XEP_0335
    xep_0352: XEP_0352
    xep_0353: XEP_0353
    xep_0359: XEP_0359
    xep_0363: XEP_0363
    xep_0369: XEP_0369
    xep_0372: XEP_0372
    xep_0377: XEP_0377
    xep_0380: XEP_0380
    xep_0382: XEP_0382
    xep_0385: XEP_0385
    xep_0394: XEP_0394
    xep_0402: XEP_0402
    xep_0403: XEP_0403
    xep_0404: XEP_0404
    xep_0405: XEP_0405
    xep_0410: XEP_0410
    xep_0421: XEP_0421
    xep_0422: XEP_0422
    xep_0424: XEP_0424
    xep_0425: XEP_0425
    xep_0428: XEP_0428
    xep_0437: XEP_0437
    xep_0439: XEP_0439
    xep_0441: XEP_0441
    xep_0444: XEP_0444
    xep_0446: XEP_0446
    xep_0447: XEP_0447
    xep_0449: XEP_0449
    xep_0454: XEP_0454
    xep_0455: XEP_0455
    xep_0461: XEP_0461
    xep_0463: XEP_0463
    xep_0469: XEP_0469
    xep_0482: XEP_0482
    xep_0490: XEP_0490
    xep_0492: XEP_0492
    xep_0494: XEP_0494
    xep_0502: XEP_0502
    xep_0511: XEP_0511
    xep_0513: XEP_0513



# Slixmpp: The Slick XMPP Library
# Copyright (C) 2011 Nathanael C. Fritz, Lance J.T. Stout
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
from slixmpp.plugins.base import register_plugin

from slixmpp.plugins.xep_0224 import stanza
from slixmpp.plugins.xep_0224.stanza import Attention
from slixmpp.plugins.xep_0224.attention import XEP_0224


register_plugin(XEP_0224)

__all__ = ['stanza', 'Attention', 'XEP_0224']




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2011  Nathanael C. Fritz
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.

from slixmpp.features.feature_mechanisms.stanza.mechanisms import Mechanisms
from slixmpp.features.feature_mechanisms.stanza.auth import Auth
from slixmpp.features.feature_mechanisms.stanza.success import Success
from slixmpp.features.feature_mechanisms.stanza.failure import Failure
from slixmpp.features.feature_mechanisms.stanza.challenge import Challenge
from slixmpp.features.feature_mechanisms.stanza.response import Response
from slixmpp.features.feature_mechanisms.stanza.abort import Abort




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2011 Nathanael C. Fritz, Lance J.T. Stout
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
import logging

from slixmpp.stanza import Message
from slixmpp.xmlstream import register_stanza_plugin
from slixmpp.xmlstream.handler import Callback
from slixmpp.xmlstream.matcher import StanzaPath
from slixmpp.plugins import BasePlugin
from slixmpp.plugins.xep_0224 import stanza


log = logging.getLogger(__name__)


class XEP_0224(BasePlugin):

    """
    XEP-0224: Attention
    """

    name = 'xep_0224'
    description = 'XEP-0224: Attention'
    dependencies = {'xep_0030'}
    stanza = stanza

    def plugin_init(self):
        """Start the XEP-0224 plugin."""
        register_stanza_plugin(Message, stanza.Attention)

        self.xmpp.register_handler(
                Callback('Attention',
                    StanzaPath('message/attention'),
                    self._handle_attention))

    def plugin_end(self):
        self.xmpp['xep_0030'].del_feature(feature=stanza.Attention.namespace)
        self.xmpp.remove_handler('Attention')

    def session_bind(self, jid):
        self.xmpp['xep_0030'].add_feature(stanza.Attention.namespace)

    def request_attention(self, to, mfrom=None, mbody=''):
        """
        Send an attention message with an optional body.

        Arguments:
            to    -- The attention request recipient's JID.
            mfrom -- Optionally specify the sender of the attention request.
            mbody -- An optional message body to include in the request.
        """
        m = self.xmpp.Message()
        m['to'] = to
        m['type'] = 'headline'
        m['attention'] = True
        if mfrom:
            m['from'] = mfrom
        m['body'] = mbody
        m.send()

    def _handle_attention(self, msg):
        """
        Raise an event after receiving a message with an attention request.

        Arguments:
            msg -- A message stanza with an attention element.
        """
        log.debug("Received attention request from: %s", msg['from'])
        self.xmpp.event('attention', msg)



# Slixmpp: The Slick XMPP Library
# Copyright (C) 2017 Emmanuel Gil Peyrot
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
from base64 import b64encode
import hashlib
import logging
from pathlib import Path

from slixmpp.plugins import BasePlugin
from slixmpp.plugins.xep_0300 import stanza, Hash


log = logging.getLogger(__name__)


class XEP_0300(BasePlugin):
    """
    XEP-0300: Use of Cryptographic Hash Functions in XMPP
    """

    name = 'xep_0300'
    description = 'XEP-0300: Use of Cryptographic Hash Functions in XMPP'
    dependencies = {'xep_0030'}
    stanza = stanza
    default_config = {
        'block_size': 1024 * 1024,  # One MiB
        'preferred': 'sha-256',
        'enable_sha-1': False,
        'enable_sha-256': True,
        'enable_sha-512': True,
        'enable_sha3-256': True,
        'enable_sha3-512': True,
        'enable_BLAKE2b256': True,
        'enable_BLAKE2b512': True,
    }

    _hashlib_function = {
        'sha-1': hashlib.sha1,
        'sha-256': hashlib.sha256,
        'sha-512': hashlib.sha512,
        'sha3-256': hashlib.sha3_256,
        'sha3-512': hashlib.sha3_512,
        'BLAKE2b256': lambda: hashlib.blake2b(digest_size=32),
        'BLAKE2b512': lambda: hashlib.blake2b(digest_size=64),
    }

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.enabled_hashes = []

    def plugin_init(self):
        namespace = 'urn:xmpp:hash-function-text-names:%s'
        self.enabled_hashes.clear()
        for algo in self._hashlib_function:
            if getattr(self, 'enable_' + algo, False):
                self.enabled_hashes.append(namespace % algo)

    def session_bind(self, jid):
        self.xmpp['xep_0030'].add_feature(Hash.namespace)

        for namespace in self.enabled_hashes:
            self.xmpp['xep_0030'].add_feature(namespace)

    def plugin_end(self):
        for namespace in self.enabled_hashes:
            self.xmpp['xep_0030'].del_feature(feature=namespace)
        self.enabled_hashes.clear()

        self.xmpp['xep_0030'].del_feature(feature=Hash.namespace)

    def compute_hash(self, filename: str | Path,
                     function: str | None = None) -> Hash:
        """
        Compute the hash of a file, and return the relevant hash XML element.

        :param filename: Path of the file to hash.
        :param function: Name of the hash function to use. If left empty,
                         the preferred function set in ``self.preferred``
                         will be used.
        :returns: A Hash element.
        """
        if function is None:
            function = self.preferred
        h = self._hashlib_function[function]()
        with open(filename, 'rb') as f:
            while True:
                block = f.read(self.block_size)
                if not block:
                    break
                h.update(block)
        hash_elem = Hash()
        hash_elem['algo'] = function
        hash_elem['value'] = b64encode(h.digest())
        return hash_elem



# Slixmpp: The Slick XMPP Library
# Copyright (C) 2020 "Maxime “pep” Buquet <pep@bouah.net>"
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
from slixmpp.plugins import register_plugin
from slixmpp.plugins.xep_0045 import stanza
from slixmpp.plugins.xep_0045.muc import XEP_0045
from slixmpp.plugins.xep_0045.stanza import MUCPresence, MUCMessage

register_plugin(XEP_0045)

__all__ = ['XEP_0045', 'MUCPresence', 'MUCMessage', 'stanza']



# Slixmpp: The Slick XMPP Library
# Copyright (C) 2011 Nathanael C. Fritz, Lance J.T. Stout
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
from slixmpp.plugins.base import register_plugin

from slixmpp.plugins.xep_0078 import stanza
from slixmpp.plugins.xep_0078.stanza import IqAuth, AuthFeature
from slixmpp.plugins.xep_0078.legacyauth import XEP_0078


register_plugin(XEP_0078)

__all__ = ['stanza', 'IqAuth', 'AuthFeature', 'XEP_0078']



# Slixmpp: The Slick XMPP Library
# Copyright (C) 2020 Mathieu Pasquet <mathieui@mathieui.net>
# This file is part of Slixmpp.
# See the file LICENSE for copying permissio
from abc import ABC
from typing import Literal

from slixmpp.stanza import Message
from slixmpp.xmlstream import (
    ElementBase,
    register_stanza_plugin,
)


NS = "urn:xmpp:fallback:0"


class Fallback(ElementBase):
    namespace = NS
    name = "fallback"
    plugin_attrib = "fallback"
    plugin_multi_attrib = "fallbacks"
    interfaces = {"for"}

    def _find_fallback(self, fallback_for: str) -> "Fallback":
        if self["for"] == fallback_for:
            return self
        for fallback in self.parent()["fallbacks"]:
            if fallback["for"] == fallback_for:
                return fallback
        raise AttributeError("No fallback for this namespace", fallback_for)

    def get_stripped_body(
        self, fallback_for: str, element: Literal["body", "subject"] = "body"
    ) -> str:
        """
        Get the body of a message, with the fallback part stripped

        :param fallback_for: namespace of the fallback to strip
        :param element: set this to "subject" get the stripped subject instead
            of body

        :return: body (or subject) content minus the fallback part
        """
        fallback = self._find_fallback(fallback_for)
        start = fallback[element]["start"]
        end = fallback[element]["end"]
        body = self.parent()[element]
        if start == end == 0:
            return ""
        if start <= end < len(body):
            return body[:start] + body[end:]
        else:
            return body


class FallbackMixin(ABC):
    namespace = NS
    name = NotImplemented
    plugin_attrib = NotImplemented
    interfaces = {"start", "end"}

    def set_start(self, v: int):
        self._set_attr("start", str(v))

    def get_start(self):
        return _int_or_zero(self._get_attr("start"))

    def set_end(self, v: int):
        self._set_attr("end", str(v))

    def get_end(self):
        return _int_or_zero(self._get_attr("end"))


class FallbackBody(FallbackMixin, ElementBase):
    name = plugin_attrib = "body"


class FallbackSubject(FallbackMixin, ElementBase):
    name = plugin_attrib = "subject"


def _int_or_zero(v: str):
    try:
        return int(v)
    except ValueError:
        return 0


def register_plugins():
    register_stanza_plugin(Message, Fallback, iterable=True)
    register_stanza_plugin(Fallback, FallbackBody)
    register_stanza_plugin(Fallback, FallbackSubject)




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2011  Nathanael C. Fritz
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
from slixmpp.plugins.base import register_plugin

from .limits import FeatureLimits


register_plugin(FeatureLimits)



# Slixmpp: The Slick XMPP Library
# Copyright (C) 2012 Nathanael C. Fritz, Lance J.T. Stout
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.

from slixmpp.xmlstream import ElementBase


class UserGaming(ElementBase):

    name = 'game'
    namespace = 'urn:xmpp:gaming:0'
    plugin_attrib = 'gaming'
    interfaces = {'character_name', 'character_profile', 'name',
                  'level', 'server_address', 'server_name', 'uri'}
    sub_interfaces = interfaces



"""
Stanza interfaces for XEP-0511: Link Metadata

Usage:

>>> register_plugin()  # automatically done if you use this plugin
>>> msg = Message()
>>> msg["link_metadata"]["about"] = "https://the.link.example.com/what-was-linked-to"
>>> msg["link_metadata"]["title"] = "A cool title"
>>> msg.pretty_print()
<message xmlns="jabber:client">
  <Description xmlns="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" rdf:about="https://the.link.example.com/what-was-linked-to">
    <title xmlns="https://ogp.me/ns#">A cool title</title>
  </Description>
</message>
"""

# Slixmpp: The Slick XMPP Library
# Copyright (C) 2012 Nathanael C. Fritz, Lance J.T. Stout
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.


from slixmpp.stanza.message import Message
from slixmpp.xmlstream import ElementBase, register_stanza_plugin

RDF_NAMESPACE = "http://www.w3.org/1999/02/22-rdf-syntax-ns#"
OG_NAMESPACE = "https://ogp.me/ns#"


class LinkMetadata(ElementBase):
    name = "Description"
    namespace = RDF_NAMESPACE
    plugin_attrib = "link_metadata"
    plugin_multi_attrib = "link_metadatas"
    interfaces = {"about", "title", "description", "url", "image", "type", "site_name"}

    def _set_og(self, el: ElementBase, value: str) -> None:
        el.xml.text = value
        self.xml.append(el.xml)

    def _get_og(self, el: type[ElementBase]) -> str | None:
        child = self.xml.find(f"{{{el.namespace}}}{el.name}")
        if child is None:
            return None
        return child.text

    def set_title(self, v: str) -> None:
        self._set_og(Title(), v)

    def get_title(self) -> str | None:
        return self._get_og(Title)

    def set_description(self, v: str) -> None:
        self._set_og(Description(), v)

    def get_description(self) -> str | None:
        return self._get_og(Description)

    def set_url(self, v: str) -> None:
        self._set_og(Url(), v)

    def get_url(self) -> str | None:
        return self._get_og(Url)

    def set_image(self, v: str) -> None:
        self._set_og(Image(), v)

    def get_image(self) -> str | None:
        return self._get_og(Image)

    def set_type(self, v: str) -> None:
        self._set_og(Type_(), v)

    def get_type(self) -> str | None:
        return self._get_og(Type_)

    def set_site_name(self, v: str) -> None:
        self._set_og(SiteName(), v)

    def get_site_name(self) -> str | None:
        return self._get_og(SiteName)

    # The following methods look overly complex, but I did not find anything better.
    # Maybe the following snippet helps understand what ElementBase does that breaks
    # the expected behaviour of get().
    # >>> msg = ET.Element("{jabber:client}message")
    # >>> desc = ET.Element("{http://www.w3.org/1999/02/22-rdf-syntax-ns#}Description")
    # >>> desc.set("{http://www.w3.org/1999/02/22-rdf-syntax-ns#}about", "link")
    # >>> msg.append(desc)
    # >>> print(ET.tostring(msg, 'unicode'))
    # <ns0:message xmlns:ns0="jabber:client" xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#">
    #   <rdf:Description rdf:about="link" />
    # </ns0:message>
    # The "rdf" prefix is automagically set because it is listed as a well-known prefix in python's stdlib `xml.etree`.

    def get_about(self) -> str | None:
        about = self.xml.get(f"{{{RDF_NAMESPACE}}}about")
        if about is not None:
            return about

        # When we set 'about' using set_about() below, the code above does not work.
        # Handling prefix definitions manually is the best I could come up with.
        for attrib, value in self.xml.attrib.items():
            if attrib.startswith("xmlns:") and value == RDF_NAMESPACE:
                prefix = attrib.removeprefix("xmlns:")
                return self.xml.get(f"{prefix}:about")

        return None

    def set_about(self, about: str) -> None:
        # If we use…
        # self.xml.set(f"{{{RDF_NAMESPACE}}}about", about)
        # …the about attrib is not visible in the output of ET.tostring().
        # It is a problem, since this is what we send in the stream.
        # It is probably due to some special handling of namespaces in ElementBase.
        # The following workaround seems to be OK…
        self.xml.set("xmlns:rdf", RDF_NAMESPACE)
        self.xml.set("rdf:about", about)


class _OpenGraphMixin(ElementBase):
    namespace = OG_NAMESPACE


class Title(_OpenGraphMixin):
    name = plugin_attrib = "title"


class Description(_OpenGraphMixin):
    name = plugin_attrib = "description"


class Url(_OpenGraphMixin):
    name = plugin_attrib = "url"


class Image(_OpenGraphMixin):
    name = plugin_attrib = "image"


class Type_(_OpenGraphMixin):
    name = plugin_attrib = "type"


class SiteName(_OpenGraphMixin):
    name = plugin_attrib = "site_name"


def register_plugin() -> None:
    for plugin in Title, Description, Url, Image, Type_, SiteName:
        register_stanza_plugin(LinkMetadata, plugin)
    register_stanza_plugin(Message, LinkMetadata, iterable=True)




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2012 Nathanael C. Fritz, Lance J.T. Stout
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
import logging

from asyncio import Future
from slixmpp.xmlstream import ElementBase
from slixmpp.plugins.base import BasePlugin, register_plugin


log = logging.getLogger(__name__)


class XEP_0223(BasePlugin):

    """
    XEP-0223: Persistent Storage of Private Data via PubSub

    If a specific pubsub node requires additional publish options, edit the
    :attr:`.node_profile` attribute of this plugin:

    .. code-block:: python

        self.xmpp.plugin["xep_0223"].node_profiles["urn:some:node"] = {
            "pubsub#max_items" = "max"
        }

    This makes :meth:`.store` add these publish options whenever it is called
    for the ``urn:some:node`` node.
    """

    name = 'xep_0223'
    description = 'XEP-0223: Persistent Storage of Private Data via PubSub'
    dependencies = {'xep_0163', 'xep_0060', 'xep_0004'}

    profile = {'pubsub#persist_items': True,
               'pubsub#access_model': 'whitelist'}
    node_profiles = dict[str, dict[str, str]]()

    def configure(self, node: str, **iqkwargs) -> Future:
        """
        Update a node's configuration to match the private storage profile.

        :param node: Node to set the configuration at.
        """
        config = self.xmpp['xep_0004'].stanza.Form()
        config['type'] = 'submit'
        config.add_field(
            var='FORM_TYPE',
            ftype='hidden',
            value='http://jabber.org/protocol/pubsub#node_config')

        for field, value in self.profile.items():
            config.add_field(var=field, value=value)

        return self.xmpp['xep_0060'].set_node_config(
            jid=None, node=node, config=config, **iqkwargs
        )

    def store(self, stanza: ElementBase, node: str | None = None,
              id: str | None = None, **pubsubkwargs) -> Future:
        """
        Store private data via PEP.

        This is just a (very) thin wrapper around the XEP-0060 publish()
        method to set the defaults expected by PEP.

        :param stanza:  The private content to store.
        :param node: The node to publish the content to. If not specified,
                     the stanza's namespace will be used.
        :param id: Optionally specify the ID of the item.
        :param options: Publish options to use, which will be modified to
                        fit the persistent storage option profile.
        """
        options = pubsubkwargs.pop('options', None)
        if not options:
            options = self.xmpp['xep_0004'].stanza.Form()
            options['type'] = 'submit'
            options.add_field(
                var='FORM_TYPE',
                ftype='hidden',
                value='http://jabber.org/protocol/pubsub#publish-options')

        fields = options['fields']
        profile = self.profile | self.node_profiles.get(node, {})
        for field, value in profile.items():
            if field not in fields:
                options.add_field(var=field)
            options.get_fields()[field]['value'] = value

        pubsubkwargs['options'] = options
        return self.xmpp['xep_0163'].publish(stanza, node, id=id, **pubsubkwargs)

    def retrieve(self, node: str, id: str | None = None,
                 item_ids: list[str] | None = None, **iqkwargs) -> Future:
        """
        Retrieve private data via PEP.

        This is just a (very) thin wrapper around the XEP-0060 publish()
        method to set the defaults expected by PEP.

        :param node: The node to retrieve content from.
        :param id: Optionally specify the ID of the item.
        :param item_ids: Specify a group of IDs. If id is also specified, it
                         will be included in item_ids.
        """
        if item_ids is None:
            item_ids = []
        if id is not None:
            item_ids.append(id)

        return self.xmpp['xep_0060'].get_items(
            jid=None, node=node,
            item_ids=item_ids,
            **iqkwargs
        )


register_plugin(XEP_0223)



from slixmpp.plugins import BasePlugin
from . import stanza


class XEP_0317(BasePlugin):
    """
    XEP-0317: Hats
    """
    name = 'xep_0317'
    description = 'XEP-0317: Hats'
    dependencies = {'xep_0030', 'xep_0045', 'xep_0050'}
    stanza = stanza
    namespace = stanza.NS

    def plugin_init(self):
        stanza.register_plugin()



from slixmpp.plugins import BasePlugin
from slixmpp.types import JidStr
from slixmpp.xmlstream import StanzaBase
from slixmpp.xmlstream.handler import Callback
from slixmpp.xmlstream.matcher import StanzaPath

from . import stanza


class XEP_0461(BasePlugin):
    """XEP-0461: Message Replies"""

    name = "xep_0461"
    description = "XEP-0461: Message Replies"

    dependencies = {"xep_0030", "xep_0428"}
    stanza = stanza
    namespace = stanza.NS

    def plugin_init(self) -> None:
        stanza.register_plugins()
        self.xmpp.register_handler(
            Callback(
                "Message replied to",
                StanzaPath("message/reply"),
                self._handle_reply_to_message,
            )
        )

    def plugin_end(self):
        self.xmpp.plugin["xep_0030"].del_feature(feature=stanza.NS)

    def session_bind(self, jid):
        self.xmpp.plugin["xep_0030"].add_feature(feature=stanza.NS)

    def _handle_reply_to_message(self, msg: StanzaBase):
        self.xmpp.event("message_reply", msg)

    def make_reply(self, reply_to: JidStr, reply_id: str,
                   fallback: str | None = None,
                   quoted_nick: str | None = None, **msg_kwargs):
        """Create a replies message stanza

        :param reply_to: Full JID of the quoted author
        :param reply_id: ID of the message to reply to
        :param fallback: Body of the quoted message
        :param quoted_nick: nickname of the quoted participant
        :param msg_kwargs: Parameters are consistent with the make_message method,
            required parameters are ``mto`` and ``mbody``
        """

        msg = self.xmpp.make_message(**msg_kwargs)
        msg["reply"]["to"] = reply_to
        msg["reply"]["id"] = reply_id
        if fallback:
            msg["reply"].add_quoted_fallback(fallback, quoted_nick)
        return msg

    def send_reply(self, reply_to: JidStr, reply_id: str,
                   fallback: str | None = None,
                   quoted_nick: str | None = None, **msg_kwargs):
        """

        :param reply_to: Full JID of the quoted author
        :param reply_id: ID of the message to reply to
        :param fallback: Body of the quoted message
        :param quoted_nick: nickname of the quoted participant
        """
        msg = self.make_reply(reply_to, reply_id, fallback, quoted_nick, **msg_kwargs)
        msg.send()




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2012 Nathanael C. Fritz, Lance J.T. Stout
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
from slixmpp.plugins.base import register_plugin

from slixmpp.plugins.xep_0027.stanza import Signed, Encrypted
from slixmpp.plugins.xep_0027.gpg import XEP_0027


register_plugin(XEP_0027)

__all__ = ['Signed', 'Encrypted', 'XEP_0027']



import logging

from slixmpp.plugins import BasePlugin

from . import stanza

log = logging.getLogger(__name__)


class XEP_0264(BasePlugin):

    """
    XEP-0264: Jingle Content Thumbnails

    Can also be used with 0385 (Stateless inline media sharing)
    """

    name = "xep_0264"
    description = "XEP-0264: Jingle Content Thumbnails"
    dependencies = {"xep_0234"}
    stanza = stanza

    def plugin_init(self):
        stanza.register_plugin()




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2010  Nathanael C. Fritz
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
from slixmpp.xmlstream.handler import Callback
from slixmpp.xmlstream.stanzabase import StanzaBase


class XMLCallback(Callback):

    """
    The XMLCallback class is identical to the normal Callback class,
    except that XML contents of matched stanzas will be processed instead
    of the stanza objects themselves.

    Methods:
        run -- Overrides Callback.run
    """

    def run(self, payload: StanzaBase, instream: bool = False) -> None:
        """
        Execute the callback function with the matched stanza's
        XML contents, instead of the stanza itself.

        Overrides BaseHandler.run

        Arguments:
            payload  -- The matched stanza object.
            instream -- Force the handler to execute during
                        stream processing. Used only by prerun.
                        Defaults to False.
        """
        Callback.run(self, payload.xml, instream)  # type: ignore



# Slixmpp: The Slick XMPP Library
# Copyright (C) 2012 Nathanael C. Fritz, Lance J.T. Stout
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
from slixmpp.plugins.base import register_plugin

from slixmpp.plugins.xep_0108 import stanza
from slixmpp.plugins.xep_0108.stanza import UserActivity
from slixmpp.plugins.xep_0108.user_activity import XEP_0108


register_plugin(XEP_0108)

__all__ = ['stanza', 'UserActivity', 'XEP_0108']




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2018 Maxime “pep” Buquet
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
from slixmpp import Message
from slixmpp.plugins import BasePlugin
from slixmpp.xmlstream import register_stanza_plugin
from slixmpp.plugins.xep_0335 import JSON_Container
from slixmpp.plugins.xep_0335 import stanza


class XEP_0335(BasePlugin):

    name = 'xep_0335'
    description = 'XEP-0335: JSON Containers'
    stanza = stanza

    def plugin_init(self):
        register_stanza_plugin(Message, JSON_Container)




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2020 Mathieu Pasquet <mathieui@mathieui.net>
# This file is part of Slixmpp.
# See the file LICENSE for copying permissio
from slixmpp.stanza import Iq
from slixmpp.stanza.roster import Roster, RosterItem
from slixmpp.xmlstream import (
    ElementBase,
    register_stanza_plugin,
)

from slixmpp.plugins.xep_0369.stanza import (
    Join,
    Leave,
)

NS = 'urn:xmpp:mix:pam:2'
NS_ROSTER = 'urn:xmpp:mix:roster:0'


class ClientJoin(ElementBase):
    namespace = NS
    name = 'client-join'
    plugin_attrib = 'client_join'
    interfaces = {'channel'}


class ClientLeave(ElementBase):
    namespace = NS
    name = 'client-leave'
    plugin_attrib = 'client_leave'
    interfaces = {'channel'}


class Annotate(ElementBase):
    namespace = NS_ROSTER
    name = 'annotate'
    plugin_attrib = 'annotate'


class Channel(ElementBase):
    namespace = NS_ROSTER
    name = 'channel'
    plugin_attrib = 'channel'
    interfaces = {'participant-id'}


def register_plugins():
    register_stanza_plugin(Iq, ClientJoin)
    register_stanza_plugin(ClientJoin, Join)

    register_stanza_plugin(Iq, ClientLeave)
    register_stanza_plugin(ClientLeave, Leave)

    register_stanza_plugin(Roster, Annotate)
    register_stanza_plugin(RosterItem, Channel)




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2020 Mathieu Pasquet <mathieui@mathieui.net>
# This file is part of Slixmpp.
# See the file LICENSE for copying permissio
from slixmpp.stanza import Presence
from slixmpp.xmlstream import (
    register_stanza_plugin,
    ElementBase,
)

from slixmpp.plugins.xep_0060.stanza import (
    Item,
    EventItem,
)


NS = 'urn:xmpp:mix:presence:0'


class MIXPresence(ElementBase):
    namespace = NS
    name = 'mix'
    plugin_attrib = 'mix'
    interfaces = {'jid', 'nick'}
    sub_interfaces = {'jid', 'nick'}


def register_plugins():
    register_stanza_plugin(Presence, MIXPresence)
    register_stanza_plugin(Item, Presence)
    register_stanza_plugin(EventItem, Presence)



# Slixmpp: The Slick XMPP Library
# Copyright (C) 2010 Nathanael C. Fritz, Lance J.T. Stout
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
from slixmpp.plugins.base import register_plugin

from slixmpp.plugins.xep_0128.static import StaticExtendedDisco
from slixmpp.plugins.xep_0128.extended_disco import XEP_0128


register_plugin(XEP_0128)

__all__ = ['StaticExtendedDisco', 'XEP_0128']



# Slixmpp: The Slick XMPP Library
# This file is part of Slixmpp
# See the file LICENSE for copying permission
import uuid
import logging

from slixmpp import JID
from slixmpp.stanza import Message, Iq
from slixmpp.exceptions import XMPPError
from slixmpp.xmlstream.handler import CoroutineCallback
from slixmpp.xmlstream.matcher import StanzaPath
from slixmpp.xmlstream import register_stanza_plugin
from slixmpp.plugins import BasePlugin
from slixmpp.plugins.xep_0047 import stanza, Open, Close, Data, IBBytestream


log = logging.getLogger(__name__)


class XEP_0047(BasePlugin):
    """
    XEP-0047: In-Band Bytestreams

    Events registered by this plugin:

    - :term:`ibb_stream_start`
    - :term:`ibb_stream_end`
    - :term:`ibb_stream_data`
    - :term:`stream:[stream id]:[peer jid]`

    Plugin Parameters:

    - ``block_size`` (default: ``4096``): default block size to negotiate
    - ``max_block_size`` (default: ``8192``): max block size to accept
    - ``auto_accept`` (default: ``False``): if incoming streams should be
        accepted automatically.

    - :term:`authorized (0047 version)`
    - :term:`authorized_sid (0047 version)`
    - :term:`preauthorize_sid (0047 version)`
    - :term:`get_stream`
    - :term:`set_stream`
    - :term:`del_stream`
    """

    name = 'xep_0047'
    description = 'XEP-0047: In-Band Bytestreams'
    dependencies = {'xep_0030'}
    stanza = stanza
    default_config = {
        'block_size': 4096,
        'max_block_size': 8192,
        'auto_accept': False,
    }

    def plugin_init(self):
        self._streams = {}
        self._preauthed_sids = {}

        register_stanza_plugin(Iq, Open)
        register_stanza_plugin(Iq, Close)
        register_stanza_plugin(Iq, Data)
        register_stanza_plugin(Message, Data)

        self.xmpp.register_handler(CoroutineCallback(
            'IBB Open',
            StanzaPath('iq@type=set/ibb_open'),
            self._handle_open_request))

        self.xmpp.register_handler(CoroutineCallback(
            'IBB Close',
            StanzaPath('iq@type=set/ibb_close'),
            self._handle_close))

        self.xmpp.register_handler(CoroutineCallback(
            'IBB Data',
            StanzaPath('iq@type=set/ibb_data'),
            self._handle_data))

        self.xmpp.register_handler(CoroutineCallback(
            'IBB Message Data',
            StanzaPath('message/ibb_data'),
            self._handle_data))

        self.api.register(self._authorized, 'authorized', default=True)
        self.api.register(self._authorized_sid, 'authorized_sid', default=True)
        self.api.register(self._preauthorize_sid, 'preauthorize_sid', default=True)
        self.api.register(self._get_stream, 'get_stream', default=True)
        self.api.register(self._set_stream, 'set_stream', default=True)
        self.api.register(self._del_stream, 'del_stream', default=True)

    def plugin_end(self):
        self.xmpp.remove_handler('IBB Open')
        self.xmpp.remove_handler('IBB Close')
        self.xmpp.remove_handler('IBB Data')
        self.xmpp.remove_handler('IBB Message Data')
        self.xmpp['xep_0030'].del_feature(feature='http://jabber.org/protocol/ibb')

    def session_bind(self, jid):
        self.xmpp['xep_0030'].add_feature('http://jabber.org/protocol/ibb')

    def _get_stream(self, jid, sid, peer_jid, data):
        return self._streams.get((jid, sid, peer_jid), None)

    def _set_stream(self, jid, sid, peer_jid, stream):
        self._streams[(jid, sid, peer_jid)] = stream

    def _del_stream(self, jid, sid, peer_jid, data):
        if (jid, sid, peer_jid) in self._streams:
            del self._streams[(jid, sid, peer_jid)]

    async def _accept_stream(self, iq):
        receiver = iq['to']
        sender = iq['from']
        sid = iq['ibb_open']['sid']

        if await self.api['authorized_sid'](receiver, sid, sender, iq):
            return True
        return await self.api['authorized'](receiver, sid, sender, iq)

    def _authorized(self, jid, sid, ifrom, iq):
        if self.auto_accept:
            return True
        return False

    def _authorized_sid(self, jid, sid, ifrom, iq):
        if (jid, sid, ifrom) in self._preauthed_sids:
            del self._preauthed_sids[(jid, sid, ifrom)]
            return True
        return False

    def _preauthorize_sid(self, jid, sid, ifrom, data):
        self._preauthed_sids[(jid, sid, ifrom)] = True

    async def open_stream(self, jid: JID, *, block_size: int | None = None,
                          sid: str | None = None, use_messages: bool = False,
                          ifrom: JID | None = None,
                          **iqkwargs) -> IBBytestream:
        """Open an IBB stream with a peer JID.

        .. versionchanged:: 1.8.0
            This function is now a coroutine and must be awaited.
            All parameters except ``jid`` are keyword-args only.

        :param jid: The remote JID to initiate the stream with.
        :param block_size: The block size to advertise.
        :param sid: The IBB stream id (if not provided, will be auto-generated).
        :param use_messages: If the stream should use message stanzas instead of iqs.
        :returns: The opened byte stream with the remote JID
        :raises .IqError: When the remote entity denied the stream.
        """
        if sid is None:
            sid = str(uuid.uuid4())
        if block_size is None:
            block_size = self.block_size

        iq = self.xmpp.make_iq_set(ito=jid, ifrom=ifrom)
        iq['ibb_open']['block_size'] = block_size
        iq['ibb_open']['sid'] = sid
        iq['ibb_open']['stanza'] = 'message' if use_messages else 'iq'

        stream = IBBytestream(self.xmpp, sid, block_size,
                              iq['from'], iq['to'], use_messages)

        callback = iqkwargs.pop('callback', None)
        result = await iq.send(**iqkwargs)

        log.debug('IBB stream (%s) accepted by %s', stream.sid, result['from'])
        stream.self_jid = result['to']
        stream.peer_jid = result['from']
        stream.stream_started = True
        await self.api['set_stream'](stream.self_jid, stream.sid, stream.peer_jid, stream)
        if callback is not None:
            self.xmpp.add_event_handler('ibb_stream_start', callback, disposable=True)
        self.xmpp.event('ibb_stream_start', stream)
        self.xmpp.event('stream:%s:%s' % (stream.sid, stream.peer_jid), stream)
        return stream

    async def _handle_open_request(self, iq: Iq):
        sid = iq['ibb_open']['sid']
        size = iq['ibb_open']['block_size'] or self.block_size

        log.debug('Received IBB stream request from %s', iq['from'])

        if not sid:
            raise XMPPError(etype='modify', condition='bad-request')

        if not await self._accept_stream(iq):
            raise XMPPError(etype='cancel', condition='not-acceptable')

        if size > self.max_block_size:
            raise XMPPError('resource-constraint')

        stream = IBBytestream(self.xmpp, sid, size,
                              iq['to'], iq['from'])
        stream.stream_started = True
        await self.api['set_stream'](stream.self_jid, stream.sid, stream.peer_jid, stream)
        iq.reply().send()

        self.xmpp.event('ibb_stream_start', stream)
        self.xmpp.event('stream:%s:%s' % (sid, stream.peer_jid), stream)

    async def _handle_data(self, stanza: Iq | Message):
        sid = stanza['ibb_data']['sid']
        stream = await self.api['get_stream'](stanza['to'], sid, stanza['from'])
        if stream is not None and stanza['from'] == stream.peer_jid:
            stream._recv_data(stanza)
        else:
            raise XMPPError('item-not-found')

    async def _handle_close(self, iq: Iq):
        sid = iq['ibb_close']['sid']
        stream = await self.api['get_stream'](iq['to'], sid, iq['from'])
        if stream is not None and iq['from'] == stream.peer_jid:
            stream._closed(iq)
            await self.api['del_stream'](stream.self_jid, stream.sid, stream.peer_jid)
        else:
            raise XMPPError('item-not-found')



import builtins
import hashlib
from sys import byteorder

from typing import Callable

bytes_ = builtins.bytes  # alias the stdlib type but ew


def unicode(text: bytes_ | str) -> str:
    if not isinstance(text, str):
        return text.decode('utf-8')
    else:
        return text


def bytes(text: str | bytes_ | None) -> bytes_:
    """
    Convert Unicode text to UTF-8 encoded bytes.

    Since Python 2.6+ and Python 3+ have similar but incompatible
    signatures, this function unifies the two to keep code sane.

    :param text: Unicode text to convert to bytes
    :rtype: bytes (Python3), str (Python2.6+)
    """
    if text is None:
        return b''

    if isinstance(text, builtins.bytes):
        # We already have bytes, so do nothing
        return text
    if isinstance(text, list):
        # Convert a list of integers to bytes
        return builtins.bytes(text)
    else:
        # Convert UTF-8 text to bytes
        return builtins.bytes(text, encoding='utf-8')


def quote(text: str | bytes_) -> bytes_:
    """
    Enclose in quotes and escape internal slashes and double quotes.

    :param text: A Unicode or byte string.
    """
    text = bytes(text)
    return b'"' + text.replace(b'\\', b'\\\\').replace(b'"', b'\\"') + b'"'


def num_to_bytes(num: int) -> bytes_:
    """
    Convert an integer into a four byte sequence.

    :param integer num: An integer to convert to its byte representation.
    """
    bval = b''
    bval += bytes(chr(0xFF & (num >> 24)))
    bval += bytes(chr(0xFF & (num >> 16)))
    bval += bytes(chr(0xFF & (num >> 8)))
    bval += bytes(chr(0xFF & (num >> 0)))
    return bval


def bytes_to_num(bval: bytes_) -> int:
    """
    Convert a four byte sequence to an integer.

    :param bytes bval: A four byte sequence to turn into an integer.
    """
    num = 0
    num += (bval[0] << 24)
    num += (bval[1] << 16)
    num += (bval[2] << 8)
    num += (bval[3])
    return num


def XOR(x: bytes_, y: bytes_) -> bytes_:
    """
    Return the results of an XOR operation on two equal length byte strings.

    :param bytes x: A byte string
    :param bytes y: A byte string
    :rtype: bytes
    """
    # This operation is faster than the previous zip() + individual xor
    # by a factor of 5
    return (
        int.from_bytes(x, byteorder) ^ int.from_bytes(y, byteorder)
    ).to_bytes(len(x), byteorder)


def hash(name: str) -> Callable | None:
    """
    Return a hash function implementing the given algorithm.

    :param name: The name of the hashing algorithm to use.
    :type name: string

    :rtype: function
    """
    name = name.lower()
    if name.startswith('sha-'):
        name = 'sha' + name[4:]
    if name in dir(hashlib):
        return getattr(hashlib, name)
    return None


def hashes() -> list[str]:
    """
    Return a list of available hashing algorithms.

    :rtype: list of strings
    """
    t = []
    if 'md5' in dir(hashlib):
        t = ['MD5']
    if 'md2' in dir(hashlib):
        t += ['MD2']
    hashes = ['SHA-' + h[3:] for h in dir(hashlib) if h.startswith('sha')]
    return t + hashes




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2011  Nathanael C. Fritz
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
from slixmpp.xmlstream import ElementBase


class Limits(ElementBase):
    name = 'limits'
    plugin_attrib = 'limits'
    namespace = 'urn:xmpp:stream-limits:0'
    interfaces = {'max-bytes', 'idle-seconds'}
    sub_interfaces = interfaces




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2012 Nathanael C. Fritz, Lance J.T. Stout
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
from slixmpp.xmlstream import ElementBase


class PrivateXML(ElementBase):

    name = 'query'
    namespace = 'jabber:iq:private'
    plugin_attrib = 'private'
    interfaces = set()




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2011  Nathanael C. Fritz
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
import logging

from slixmpp.stanza import Iq, StreamFeatures
from slixmpp.xmlstream import register_stanza_plugin
from slixmpp.plugins import BasePlugin

from slixmpp.features.feature_session import stanza
from typing import  ClassVar


log = logging.getLogger(__name__)


class FeatureSession(BasePlugin):

    name = 'feature_session'
    description = 'RFC 3920: Stream Feature: Start Session'
    dependencies: ClassVar[set[str]] = set()
    stanza = stanza

    def plugin_init(self):
        self.xmpp.register_feature('session',
                self._handle_start_session,
                restart=False,
                order=10001)

        register_stanza_plugin(Iq, stanza.Session)
        register_stanza_plugin(StreamFeatures, stanza.Session)

    async def _handle_start_session(self, features):
        """
        Handle the start of the session.

        Arguments:
            feature -- The stream features element.
        """
        if features['session']['optional']:
            self.xmpp.sessionstarted = True
            self.xmpp.event('session_start')
            return

        iq = self.xmpp.Iq()
        iq['type'] = 'set'
        iq.enable('session')
        await iq.send(callback=self._on_start_session_response)

    def _on_start_session_response(self, response):
        self.xmpp.features.add('session')

        log.debug("Established Session")
        self.xmpp.sessionstarted = True
        self.xmpp.event('session_start')



# Slixmpp: The Slick XMPP Library
# Copyright (C) 2012 Nathanael C. Fritz, Lance J.T. Stout
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
from slixmpp.plugins.base import register_plugin

from slixmpp.plugins.xep_0352.stanza import Active, Inactive, ClientStateIndication
from slixmpp.plugins.xep_0352.csi import XEP_0352


register_plugin(XEP_0352)

__all__ = ['Active', 'Inactive', 'ClientStateIndication', 'XEP_0352']




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2011 Nathanael C. Fritz
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
from slixmpp.xmlstream import ElementBase


class IqAuth(ElementBase):
    namespace = 'jabber:iq:auth'
    name = 'query'
    plugin_attrib = 'auth'
    interfaces = {'fields', 'username', 'password', 'resource', 'digest'}
    sub_interfaces = {'username', 'password', 'resource', 'digest'}
    plugin_tag_map = {}
    plugin_attrib_map = {}

    def get_fields(self):
        fields = set()
        for field in self.sub_interfaces:
            if self.xml.find('{%s}%s' % (self.namespace, field)) is not None:
                fields.add(field)
        return fields

    def set_resource(self, value):
        self._set_sub_text('resource', value, keep=True)

    def set_password(self, value):
        self._set_sub_text('password', value, keep=True)


class AuthFeature(ElementBase):
    namespace = 'http://jabber.org/features/iq-auth'
    name = 'auth'
    plugin_attrib = 'auth'
    interfaces = set()
    plugin_tag_map = {}
    plugin_attrib_map = {}



# Slixmpp: The Slick XMPP Library
# Copyright (C) 2010  Nathanael C. Fritz
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.

import warnings
from importlib.metadata import PackageNotFoundError, version

__version_info__ = (0, 0, 0)

def warn_pride():
    warnings.warn(
        f"Version '{__version__}' does not respect the pride versioning spec."
    )

try:
    __version__ = version("slixmpp")
except PackageNotFoundError:
    warnings.warn("Cannot set the version because the 'slixmpp' package is not installed.")
    __version__ = "not-installed"
else:
    try:
        ver = tuple(int(x) for x in __version__.split(".")[:3])
    except ValueError:
        warn_pride()
    else:
        if len(ver) == 3:
            __version_info__ = ver
        else:
            warn_pride()




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2013 Nathanael C. Fritz, Lance J.T. Stout
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
from slixmpp.xmlstream import ElementBase


class SI(ElementBase):
    name = 'si'
    namespace = 'http://jabber.org/protocol/si'
    plugin_attrib = 'si'
    interfaces = {'id', 'mime_type', 'profile'}

    def get_mime_type(self):
        return self._get_attr('mime-type', 'application/octet-stream')

    def set_mime_type(self, value):
        self._set_attr('mime-type', value)

    def del_mime_type(self):
        self._del_attr('mime-type')



# Slixmpp: The Slick XMPP Library
# Copyright (C) 2010 Nathanael C. Fritz
# Copyright (C) 2020 "Maxime “pep” Buquet <pep@bouah.net>"
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.

import asyncio
import logging
from collections import defaultdict
from datetime import datetime
from typing import (
    Any,
)

from slixmpp import (
    Presence,
    Message,
    Iq,
    JID,
)
from slixmpp.plugins import BasePlugin
from slixmpp.xmlstream import register_stanza_plugin
from slixmpp.xmlstream.handler.callback import Callback
from slixmpp.xmlstream.matcher.stanzapath import StanzaPath
from slixmpp.xmlstream.matcher.xmlmask import MatchXMLMask
from slixmpp.exceptions import PresenceError

from slixmpp.plugins.xep_0004 import Form
from slixmpp.plugins.xep_0045 import stanza
from slixmpp.plugins.xep_0045.stanza import (
    MUCInvite,
    MUCDecline,
    MUCDestroy,
    MUCPresence,
    MUCJoin,
    MUCMessage,
    MUCAdminQuery,
    MUCAdminItem,
    MUCHistory,
    MUCOwnerQuery,
    MUCOwnerDestroy,
    MUCStatus,
    MUCActor,
    MUCUserItem,
)
from slixmpp.types import (
    JidStr,
    MucRole,
    MucAffiliation,
    MucRoomItem,
    MucRoomItemKeys,
    PresenceArgs,
    PresenceShows,
)

JoinResult = tuple[Presence, Message, list[Presence], list[Message]]

log = logging.getLogger(__name__)

AFFILIATIONS = ('outcast', 'member', 'admin', 'owner', 'none')
ROLES = ('moderator', 'participant', 'visitor', 'none')


class BadPfromUsage(Exception):
    """Exception raised when the pfrom parameter usage is wrong"""
    def __init__(self, is_set: bool):
        if is_set:
            message = (
                '"pfrom" parameter is set, but self.multi_from is False.'
                'This is likely an error.'
            )
        else:
            message = (
                '"pfrom" parameter is not set, but self.multi_from is True.'
                'This is likely an error.'
            )
        super().__init__(message)


class XEP_0045(BasePlugin):

    """
    XEP-0045 Multi-User Chat.


    This XEP is made for use by *clients* or components acting as clients.
    If a component is made to act as MUC service, it should only care about
    the events and the stanza plugins.

    A single component may want to simulate joins as multiple entities, in
    which case it should enable the multi_from config option.


    """

    name = 'xep_0045'
    description = 'XEP-0045: Multi-User Chat'
    dependencies = {'xep_0004', 'xep_0030', 'xep_0172', 'xep_0203'}
    stanza = stanza
    default_config = {
        'multi_from': False,
    }

    rooms: dict[JID | None, dict[JID, dict[str, MucRoomItem]]]
    our_nicks: dict[JID | None, dict[JID, str]]

    def plugin_init(self):
        self.rooms = defaultdict(lambda: defaultdict())
        self.our_nicks = defaultdict(lambda: defaultdict())
        # load MUC support in presence stanzas
        register_stanza_plugin(MUCMessage, MUCUserItem)
        register_stanza_plugin(MUCPresence, MUCUserItem)
        register_stanza_plugin(MUCUserItem, MUCActor)
        register_stanza_plugin(MUCMessage, MUCInvite)
        register_stanza_plugin(MUCMessage, MUCDecline)
        register_stanza_plugin(MUCMessage, MUCStatus)
        register_stanza_plugin(MUCPresence, MUCStatus)
        register_stanza_plugin(Presence, MUCPresence)
        register_stanza_plugin(MUCPresence, MUCDestroy)
        register_stanza_plugin(Presence, MUCJoin)
        register_stanza_plugin(MUCJoin, MUCHistory)
        register_stanza_plugin(Message, MUCMessage)
        register_stanza_plugin(Iq, MUCAdminQuery)
        register_stanza_plugin(Iq, MUCOwnerQuery)
        register_stanza_plugin(MUCOwnerQuery, MUCOwnerDestroy)
        register_stanza_plugin(MUCOwnerQuery, Form)
        register_stanza_plugin(MUCAdminQuery, MUCAdminItem, iterable=True)

        # Register handlers
        self.xmpp.register_handler(
            Callback(
                'MUCPresence',
                StanzaPath("presence/muc"),
                self._handle_groupchat_presence,
        ))
        # <x xmlns="http://jabber.org/protocol/muc"/> is only used in
        # presence when joining on the client side, and for errors on
        # the server side.
        if self.xmpp.is_component:
            self.xmpp.register_handler(
                Callback(
                    'MUCPresenceJoin',
                    StanzaPath("presence/muc_join"),
                    self._handle_groupchat_join,
            ))
        self.xmpp.register_handler(
            Callback(
                "MUCPresenceError",
                StanzaPath("presence@type=error/muc_join"),
                self._handle_presence_error,
            )
        )

        self.xmpp.register_handler(
            Callback(
                'MUCError',
                MatchXMLMask("<message xmlns='%s' type='error'><error/></message>" % self.xmpp.default_ns),
                self._handle_groupchat_error_message
        ))
        self.xmpp.register_handler(
            Callback(
                'MUCMessage',
                MatchXMLMask("<message xmlns='%s' type='groupchat'><body/></message>" % self.xmpp.default_ns),
                self._handle_groupchat_message
        ))
        self.xmpp.register_handler(
            Callback(
                'MUCSubject',
                MatchXMLMask("<message xmlns='%s' type='groupchat'><subject/></message>" % self.xmpp.default_ns),
                self._handle_groupchat_subject
        ))
        self.xmpp.register_handler(
            Callback(
                'MUCConfig',
                StanzaPath('message/muc/status'),
                self._handle_config_change
        ))
        self.xmpp.register_handler(
            Callback(
                'MUCInvite',
                StanzaPath('message/muc/invite'),
                self._handle_groupchat_invite
        ))
        self.xmpp.register_handler(
            Callback(
                'MUCDecline',
                StanzaPath('message/muc/decline'),
                self._handle_groupchat_decline
        ))
        self.xmpp.register_handler(
            Callback(
                'MUCAffiliationChange',
                StanzaPath('message/muc/item'),
                self._handle_groupchat_affiliation_change
        ))
        if not self.xmpp.is_component:
            self.multi_from = False

    def plugin_end(self):
        self.xmpp.plugin['xep_0030'].del_feature(feature=stanza.NS)

    def session_bind(self, jid):
        self.xmpp.plugin['xep_0030'].add_feature(stanza.NS)

    def _handle_groupchat_invite(self, inv: Message):
        """ Handle an invite into a muc. """
        if self.xmpp.is_component:
            self.xmpp.event('groupchat_invite', inv)
        else:
            if inv['from'] not in self.rooms[None].keys():
                self.xmpp.event("groupchat_invite", inv)

    def _handle_groupchat_decline(self, decl: Message):
        """Handle an invitation decline."""
        if self.xmpp.is_component:
            self.xmpp.event('groupchat_decline', decl)
        else:
            if decl['from'] in self.rooms.keys():
                self.xmpp.event('groupchat_decline', decl)

    def _handle_config_change(self, msg: Message):
        """Handle a MUC configuration change (with status code)."""
        self.xmpp.event('groupchat_config_status', msg)
        self.xmpp.event('muc::%s::config_status' % msg['from'].bare, msg)

    def _handle_groupchat_affiliation_change(self, msg: Message) -> None:
        if msg.get_from().resource:
            return
        self.xmpp.event('groupchat_affiliation_change', msg)
        self.xmpp.event('muc::%s::affiliation_change' % msg['from'].bare, msg)

    def _handle_presence(self, pr: Presence):
        """As a client, handle a presence stanza"""
        got_offline = False
        got_online = False
        if self.multi_from:
            rooms = self.rooms[pr['to']]
        else:
            rooms = self.rooms[None]
        if pr['muc']['room'] not in rooms.keys():
            return
        self.xmpp.roster[pr['from']].ignore_updates = True
        entry = pr['muc'].get_stanza_values()
        entry['show'] = pr['show'] if pr['show'] in pr.showtypes else None
        entry['status'] = pr['status']

        alt_nick_element = pr.get_plugin('nick', check=True)
        entry['alt_nick'] = alt_nick_element['nick'] if alt_nick_element else ''
        if pr['type'] == 'unavailable':
            if entry['nick'] in rooms[entry['room']]:
                del rooms[entry['room']][entry['nick']]
            got_offline = True
        else:
            if entry['nick'] not in rooms[entry['room']]:
                got_online = True
            rooms[entry['room']][entry['nick']] = entry
        log.debug("MUC presence from %s/%s : %s", entry['room'], entry['nick'], entry)
        status_codes = pr['muc']['status_codes']
        if 110 in status_codes:
            if 303 in status_codes:
                if self.multi_from:
                    pto = pr['to']
                else:
                    pto = None
                self.our_nicks[pto][pr['from'].bare] = pr['muc']['item_nick']
            self.xmpp.event("muc::%s::self-presence" % entry['room'], pr)
        self.xmpp.event("muc::%s::presence" % entry['room'], pr)
        if got_offline:
            self.xmpp.event("muc::%s::got_offline" % entry['room'], pr)
        if got_online:
            self.xmpp.event("muc::%s::got_online" % entry['room'], pr)

    def _handle_presence_error(self, pr: Presence):
        """Generate MUC presence error events"""
        self.xmpp.event("muc::%s::presence-error" % pr['from'].bare, pr)

    def _handle_groupchat_presence(self, pr: Presence):
        """ Handle a presence in a muc."""
        self.xmpp.event("groupchat_presence", pr)
        self._handle_presence(pr)

    def _handle_groupchat_join(self, pr: Presence):
        """Received a join presence (as a component)"""
        self.xmpp.event('groupchat_join', pr)

    def _handle_groupchat_message(self, msg: Message):
        """ Handle a message event in a muc.
        """
        self.xmpp.event('groupchat_message', msg)
        self.xmpp.event("muc::%s::message" % msg['from'].bare, msg)

    def _handle_groupchat_error_message(self, msg: Message):
        """ Handle a message error event in a muc.
        """
        self.xmpp.event('groupchat_message_error', msg)
        self.xmpp.event("muc::%s::message_error" % msg['from'].bare, msg)

    def _handle_groupchat_subject(self, msg: Message):
        """ Handle a message coming from a muc indicating
        a change of subject (or announcing it when joining the room)
        """
        # See poezio#3452. A message containing subject _and_ (body or thread)
        # is not a subject change.
        if msg['body'] or msg['thread']:
            return
        self.xmpp.event('groupchat_subject', msg)
        self.xmpp.event('muc::%s::groupchat_subject' % msg['from'].bare, msg)

    def _raise_wrong_pfrom_usage(self, pfrom: JID | None) -> None:
        """Signal that there is a programming error in pfrom
        usage."""
        # If a client or a component acting like a client sets a pfrom=,
        # it is certainly an error.
        if not self.multi_from and pfrom is not None:
            raise BadPfromUsage(is_set=True)
        # If a component acting like multiple clients does not set a pfrom, it
        # is also certainly an error.
        elif self.multi_from and pfrom is None:
            raise BadPfromUsage(is_set=False)

    def make_join_stanza(self, room: JID, nick: str, *,
                         password: str | None = None,
                         maxchars: int | None = None,
                         maxstanzas: int | None = None,
                         seconds: int | None = None,
                         since: datetime | None = None,
                         presence_options: PresenceArgs | None = None) -> Presence:
        """
        Build the stanza for the MUC join, without sending it.

        Only one of {maxchars, maxstanzas, seconds, since} will be used, in
        that order.

        .. versionadded:: 1.9.0

        :param room: Room JID
        :param nick: User nickname in the room
        :param password: The optional room password.
        :param maxchars: Max number of characters to return from history.
        :param maxstanzas: Max number of stanzas to return from history.
        :param seconds: Fetch history until that many seconds in the past.
        :param since: Fetch history since that timestamp.
        :return: The presence stanza to send
        """
        if presence_options is None:
            presence_options = {}
        elif presence_options.get('type') == 'unavailable':
            del presence_options['type']
        pto = JID(room)
        pto.resource = nick
        stanza = self.xmpp.make_presence(
            pto=pto,
            **presence_options
        )
        stanza.enable('muc_join')
        if password is not None:
            stanza['muc_join']['password'] = password
        if maxchars is not None:
            stanza['muc_join']['history']['maxchars'] = str(maxchars)
        elif maxstanzas is not None:
            stanza['muc_join']['history']['maxstanzas'] = str(maxstanzas)
        elif seconds is not None:
            stanza['muc_join']['history']['seconds'] = str(seconds)
        elif since is not None:
            fmt = self.xmpp.plugin['xep_0082'].format_datetime(since)
            stanza['muc_join']['history']['since'] = fmt
        return stanza

    async def join_muc_wait(self, room: JID, nick: str, *,
                            password: str | None = None,
                            maxchars: int | None = None,
                            maxstanzas: int | None = None,
                            seconds: int | None = None,
                            since: datetime | None = None,
                            presence_options: PresenceArgs | None = None,
                            timeout: float = 300) -> JoinResult:
        """
        Try to join a MUC and block until we are joined or get an error.

        Only one of {maxchars, maxstanzas, seconds, since} will be used, in
        that order.

        .. versionadded:: 1.8.0

        :param room: Room JID
        :param nick: User nickname in the room
        :param password: The optional room password.
        :param maxchars: Max number of characters to return from history.
        :param maxstanzas: Max number of stanzas to return from history.
        :param seconds: Fetch history until that many seconds in the past.
        :param since: Fetch history since that timestamp.
        :param timeout: Timeout in seconds after which a TimeoutError is raised.
                        None means no timeout.
        :raises: A slixmpp.exceptions.PresenceError if the MUC returns a
                 presence error.
        :raises: An asyncio.TimeoutError if there is neither success nor
                presence error when the timeout is reached.
        :return: A tuple containing our own presence, the subject, a list
                 of occupants and a list of history messages.
        """
        if self.xmpp.is_component and not presence_options.get('pfrom'):
            raise ValueError('Components must always set the pfrom= attribute.')

        stanza = self.make_join_stanza(
            room=room,
            nick=nick,
            password=password,
            maxchars=maxchars,
            maxstanzas=maxstanzas,
            seconds=seconds,
            since=since,
            presence_options=presence_options,
        )
        if self.multi_from:
            pfrom = presence_options['pfrom']
        else:
            pfrom = None
        self.rooms[pfrom][room] = {}
        self.our_nicks[pfrom][room] = nick
        return await self._await_join(room, stanza, timeout)

    async def _await_join(self, room: JID, stanza: Presence, timeout: float = 300) -> JoinResult:
        """Do the heavy lifting for awaiting a MUC join

        A muc join, once the join stanza is sent, is:
            occupant presences → self-presence → room history → room subject
        """
        presence_done: asyncio.Future = asyncio.Future(loop=self.xmpp.loop)
        topic_received: asyncio.Future = asyncio.Future(loop=self.xmpp.loop)
        history_buffer: list[Message] = []
        occupant_buffer: list[Presence] = []

        pfrom = stanza['from'] or None

        def add_message(msg: Message):
            if pfrom and pfrom != msg['to']:
                return
            delay = msg.get_plugin('delay', check=True)
            if delay is not None and delay['from'] == room:
                history_buffer.append(msg)

        def add_occupant(pres: Presence):
            if pfrom and pfrom != pres['to']:
                return
            occupant_buffer.append(pres)

        def set_topic(msg: Message):
            if pfrom and pfrom != msg['to']:
                return
            if not topic_received.done():
                topic_received.set_result(msg)

        def set_self_presence(pres: Presence):
            if pfrom and pfrom != pres['to']:
                return
            if not presence_done.done():
                presence_done.set_result(pres)

        catch_occupants = self.xmpp.event_handler(
            "muc::%s::got_online" % room,
            add_occupant
        )
        catch_history = self.xmpp.event_handler(
            "muc::%s::message" % room,
            add_message
        )
        subject_handler = self.xmpp.event_handler(
            "muc::%s::groupchat_subject" % room,
            set_topic
        )
        self_presence = self.xmpp.event_handler(
            "muc::%s::self-presence" % room,
            set_self_presence
        )
        presence_error = self.xmpp.event_handler(
            "muc::%s::presence-error" % room,
            set_self_presence
        )

        with subject_handler, catch_history, catch_occupants:
            with self_presence, presence_error:
                stanza.send()
                done, pending = await asyncio.wait(
                    [presence_done],
                    timeout=timeout,
                )
            if pending:
                raise asyncio.TimeoutError()
            pres: Presence = presence_done.result()
            if pres['type'] == 'error':
                raise PresenceError(pres)
            done, pending = await asyncio.wait(
                [topic_received],
                timeout=timeout,
            )
            if pending:
                raise asyncio.TimeoutError()
        subject: Message = topic_received.result()
        # update known nick in case it has changed
        self.our_nicks[pfrom][room] = pres['from'].resource
        return (pres, subject, occupant_buffer, history_buffer)

    def join_muc(self, room: JID, nick: str, maxhistory: str = "0",
                 password: str = '', pstatus: str = '',
                 pshow: PresenceShows = 'chat', pfrom: JidStr = ''
                 ) -> asyncio.Future:
        """ Join the specified room, requesting 'maxhistory' lines of history.

        .. deprecated:: 1.8.0

            :meth:`join_muc_wait` will replace this old API starting from version
            1.9.0.

        """
        presence_options = PresenceArgs(
            pshow=pshow,
            pstatus=pstatus,
            pfrom=pfrom,
        )
        maxchars, maxstanzas = None, None
        if maxhistory:
            if maxhistory == "0":
                maxchars = 9
            else:
                maxstanzas = int(maxhistory)
        return asyncio.ensure_future(
            self.join_muc_wait(
                room=room,
                nick=nick,
                password=password,
                presence_options=presence_options,
                maxchars=maxchars,
                maxstanzas=maxstanzas,
            ),
            loop=self.xmpp.loop,
        )

    def leave_muc(self, room: JID, nick: str, msg: str = '', pfrom: JID | None = None):
        """ Leave the specified room.

        :param room: Room to leave.
        :param nick: Your nickname.
        :param msg: Presence status to use.

        :raises: KeyError if the room is not in our room list.
        """
        if msg:
            self.xmpp.send_presence(
                pshow='unavailable',
                pto="%s/%s" % (room, nick),
                pstatus=msg,
                pfrom=pfrom
            )
        else:
            self.xmpp.send_presence(
                pshow='unavailable',
                pto="%s/%s" % (room, nick),
                pfrom=pfrom
            )
        if pfrom in self.rooms and room in self.rooms[pfrom]:
            del self.rooms[pfrom][room]
        else:
            raise KeyError(
                f'Unable to find the room {room} in the currently joined rooms'
                + (f' for {pfrom}' if pfrom else '')
            )

    def set_subject(self, room: JidStr, subject: str, *, mfrom: JID | None = None):
        """Set a room’s subject.

        :param room: JID of the room.
        :param subject: Room subject to set.
        """
        msg = self.xmpp.make_message(room, mfrom=mfrom)
        msg['type'] = 'groupchat'
        msg['subject'] = subject
        msg.send()

    async def get_room_config(self, room: JidStr, ifrom: JID | None = None,
                              **iqkwargs) -> Form:
        """Get the room config form in 0004 plugin format.

        :param room: Room to get the config form from.
        :raises ValueError: When the form is not found.
        :returns: A form object.
        """
        iq = self.xmpp.make_iq_get(stanza.NS_OWNER, ito=room, ifrom=ifrom)
        result = await iq.send(**iqkwargs)
        form = result['mucowner_query'].get_plugin('form', check=True)
        if form is None:
            raise ValueError("Configuration form not found")
        return form

    async def set_room_config(self, room: JidStr, config: Form, *,
                              ifrom: JID | None = None, **iqkwargs):
        """Send a room config form.

        :param room: Room to send the form to.
        :param config: A filled room form.
        """
        query = MUCOwnerQuery()
        config['type'] = 'submit'
        query.append(config)
        iq = self.xmpp.make_iq_set(query, ito=room, ifrom=ifrom)
        await iq.send(**iqkwargs)

    async def cancel_config(self, room: JidStr, *,
                            ifrom: JidStr | None = None, **iqkwargs):
        """Cancel a requested config form.

        :param room: Room to cancel the form for.
        """
        query = MUCOwnerQuery()
        query['form']['type'] = 'cancel'
        iq = self.xmpp.make_iq_set(query, ito=room, ifrom=ifrom)
        await iq.send(**iqkwargs)

    async def destroy(self, room: JidStr, reason: str = '', altroom: JidStr | None = None, *,
                      ifrom: JidStr | None = None, **iqkwargs):
        """Destroy a room.

        :param room: Room JID to destroy.
        :param reason: Reason for destroying the room.
        :param altroom: An alternate room that users should join.
        """
        iq = self.xmpp.make_iq_set(ifrom=ifrom, ito=room)
        iq.enable('mucowner_query')
        iq['mucowner_query'].enable('destroy')
        if altroom:
            iq['mucowner_query']['destroy']['jid'] = altroom
        if reason:
            iq['mucowner_query']['destroy']['reason'] = reason
        await iq.send(**iqkwargs)

    async def set_affiliation(self, room: JidStr, affiliation: MucAffiliation, *,
                              jid: JidStr | None = None,
                              nick: str | None = None, reason: str = '',
                              ifrom: JidStr | None = None, **iqkwargs):
        """ Change room affiliation for a JID or nickname.

        :param room: Room to modify.
        :param affiliation: Affiliation to set.
        :param jid: User JID to use in the set operation.
        :param reason: Reason for the affiliation change.
        """
        if affiliation not in AFFILIATIONS:
            raise ValueError('%s is not a valid affiliation' % affiliation)
        if affiliation == 'outcast' and not jid:
            raise ValueError('Outcast affiliation requires using a jid')
        if not any((jid, nick)):
            raise ValueError('One of jid or nick must be set')
        iq = self.xmpp.make_iq_set(ito=room, ifrom=ifrom)
        iq['mucadmin_query']['item']['affiliation'] = affiliation
        if nick:
            iq['mucadmin_query']['item']['nick'] = nick
        if jid:
            iq['mucadmin_query']['item']['jid'] = jid
        if reason:
            iq['mucadmin_query']['item']['reason'] = reason
        await iq.send(**iqkwargs)

    async def get_affiliation_list(self, room: JidStr, affiliation: MucAffiliation, *,
                                   ifrom: JidStr | None = None, **iqkwargs) -> list[JID]:
        """Get a list of JIDs with the specified affiliation

        :param room: Room to get affiliations from.
        :param affiliation: The affiliation to list.
        """
        iq = self.xmpp.make_iq_get(stanza.NS_ADMIN, ito=room, ifrom=ifrom)
        iq['mucadmin_query']['item']['affiliation'] = affiliation
        result = await iq.send(**iqkwargs)
        return [item['jid'] for item in result['mucadmin_query']]

    async def send_affiliation_list(self, room: JidStr,
                                    affiliations: list[tuple[JidStr, MucAffiliation]], *,
                                    ifrom: JidStr | None = None, **iqkwargs):
        """Send an affiliation delta list.

        :param room: Room to send the affiliations to.
        :param affiliations: List of couples (jid, affiliation) to set.
        """
        iq = self.xmpp.make_iq_set(ito=room, ifrom=ifrom)
        for jid, affiliation in affiliations:
            item = MUCAdminItem()
            item['jid'] = jid
            item['affiliation'] = affiliation
            iq['mucadmin_query'].append(item)
        await iq.send(**iqkwargs)

    async def set_role(self, room: JidStr, nick: str, role: MucRole, *,
                       reason: str = '', ifrom: JidStr | None = None, **iqkwargs):
        """
        Change role property of a nick in a room.
        Typically, roles are temporary (they last only as long as you are in
        the room), whereas affiliations are permanent (they last across
        groupchat sessions).

        :param room: Room to modify.
        :param nick: User nickname to use in the set operation.
        :param role: Role to set.
        :param reason: Reason for the role change.
        """
        if role not in ROLES:
            raise ValueError("Role %s does not exist" % role)
        iq = self.xmpp.make_iq_set(ito=room, ifrom=ifrom)
        iq['mucadmin_query']['item']['role'] = role
        iq['mucadmin_query']['item']['nick'] = nick
        if reason:
            iq['mucadmin_query']['item']['reason'] = reason
        await iq.send(**iqkwargs)

    async def get_roles_list(self, room: JidStr, role: MucRole, *,
                             ifrom: JidStr | None = None, **iqkwargs) -> list[str]:
        """"Get a list of JIDs with the specified role

        :param room: Room to get roles from.
        :param role: The role to list.
        """
        iq = self.xmpp.make_iq_get(stanza.NS_ADMIN, ito=room, ifrom=ifrom)
        iq['mucadmin_query']['item']['role'] = role
        result = await iq.send(**iqkwargs)
        return [item['nick'] for item in result['mucadmin_query']]

    async def send_role_list(self, room: JidStr, roles: list[tuple[str, MucRole]], *,
                             ifrom: JidStr | None = None, **iqkwargs):
        """Send a role delta list.

        :param room: Room to send the roles to.
        :param roles: List of couples (nick, role) to set.
        """
        iq = self.xmpp.make_iq_set(ito=room, ifrom=ifrom)
        for nick, affiliation in roles:
            item = MUCAdminItem()
            item['nick'] = nick
            item['affiliation'] = affiliation
            iq['mucadmin_query'].append(item)
        await iq.send(**iqkwargs)

    def invite(self, room: JidStr, jid: JidStr, reason: str = '', *,
               mfrom: JidStr | None = None):
        """ Invite a jid to a room (mediated invitation).

        :param room: Room to invite the user in.
        :param jid: JID of the user to invite.
        :param reason: Reason for inviting the user.
        """
        msg = self.xmpp.make_message(room, mfrom=mfrom)
        msg['muc']['invite']['to'] = jid
        if reason:
            msg['muc']['invite']['reason'] = reason
        self.xmpp.send(msg)

    def invite_server(self, room: JidStr, jid: JidStr,
                         invite_from: JidStr, reason: str = ''):
        """Send a mediated invite to a user, as a MUC service.

        .. versionadded:: 1.8.0

        :param room: Room to invite the user in.
        :param jid: JID of the user to invite.
        :param invite_from: JID of the user to send the invitation from.
        :param reason: Reason for inviting the user.
        """
        if not self.xmpp.is_component:
            raise ValueError("Cannot use this method as a client.")
        msg = self.xmpp.make_message(jid, mfrom=room)
        msg['muc']['invite']['from'] = invite_from
        if reason:
            msg['muc']['invite']['reason'] = reason
        msg.send()

    def decline(self, room: JidStr, jid: JidStr, reason: str = '', *,
                mfrom: JidStr | None = None):
        """Decline a mediated invitation.

        :param room: Room the invitation came from.
        :param jid: JID of the user who sent the invitation.
        :param reason: Reason for declining.
        """
        msg = self.xmpp.make_message(room, mfrom=mfrom)
        msg['muc']['decline']['to'] = jid
        if reason:
            msg['muc']['decline']['reason'] = reason
        self.xmpp.send(msg)

    def request_voice(self, room: JidStr, role: str, *, mfrom: JidStr | None = None):
        """Request voice in a moderated room.

        :param room: Room to request voice from.
        """
        msg = self.xmpp.make_message(room, mfrom=mfrom)
        form = msg['form']
        form['type'] = 'submit'
        form.add_field(
            var='FORM_TYPE',
            ftype='hidden',
            value='http://jabber.org/protocol/muc#request',
        )
        form.add_field(var='muc#role', ftype='list-single',
                       label='Requested role', value=role)
        self.xmpp.send(msg)

    async def set_self_nick(self, room: JID, new_nick: str,
                            timeout: float = 60,
                            presence_options: PresenceArgs | None = None) -> str:
        """
        Set your nickname in a room.
        The room can arbitrarily decide on another nickname, so this function
        waits for the response and returns the final nickname.

        .. versionadded:: 1.9.0

        :param room: Room in which to change our nickname
        :param new_nick: Our new nickname in the room
        :param timeout: Time in seconds to wait for our new nickname response
        :param pfrom: (for components) the JID to send our presence from
        """
        new_jid = JID(room)
        new_jid.resource = new_nick
        if presence_options is None:
            presence_options = {}
        future = asyncio.Future(loop=self.xmpp.loop)

        def nickname_set(presence):
            codes = presence['muc']['status_codes']
            if 110 in codes and 303 in codes:
                future.set_result(presence)

        handler = self.xmpp.event_handler(
            f"muc::{room}::self-presence",
            nickname_set,
        )
        with handler:
            self.xmpp.make_presence(pto=new_jid, **presence_options).send()
            done, pending = await asyncio.wait([future], timeout=timeout)
        if pending:
            raise TimeoutError("Timed out waiting for server answer")

        presence = future.result()
        new_nick = presence['muc']['item_nick']
        return new_nick

    def jid_in_room(self, room: JID, jid: JID, pfrom: JID | None = None) -> bool:
        """Check if a JID is present in a room.

        :param room: Room to check.
        :param jid: full JID to check.
        """
        self._raise_wrong_pfrom_usage(pfrom)
        if not self.multi_from:
            pfrom = None
        bare_match = False
        rooms = self.rooms.get(pfrom, {})
        room = rooms.get(room, {})
        for nick in room:
            entry = room[nick]
            if not entry.get('jid'):
                continue

            if entry['jid'] == jid.full:
                return True
            elif JID(entry['jid']).bare == jid.bare:
                bare_match = True

        if bare_match:
            logging.info(
                "Could not retrieve full JID, falling back to bare JID for %s in %s",
                jid, room
            )
        return bare_match

    def get_nick(self, room: JID, jid: JID, pfrom: JID | None = None) -> str | None:
        """Get the nickname of a specific JID in a room.

        :param room: Room to inspect.
        :param jid: FULL JID whose nick to return.
        """
        self._raise_wrong_pfrom_usage(pfrom)
        if not self.multi_from:
            pfrom = None
        bare_match = None
        rooms = self.rooms.get(pfrom, {})
        room = rooms.get(room, {})
        for nick in room:
            entry = room[nick]
            if not entry.get('jid'):
                continue

            if entry['jid'] == jid.full:
                return nick
            elif JID(entry['jid']).bare == jid.bare:
                bare_match = nick

        if bare_match:
            logging.info(
                "Could not retrieve full JID, falling back to bare JID for %s in %s",
                jid, room
            )
        return bare_match

    def get_joined_rooms(self, pfrom: JID | None = None) -> list[JID]:
        """Get the list of rooms we sent a join presence to
        and did not explicitly leave.
        """
        self._raise_wrong_pfrom_usage(pfrom)
        if not self.multi_from:
            pfrom = None
        return list(self.rooms.get(pfrom, {}).keys())

    def get_our_jid_in_room(self, room_jid: JID, pfrom: JID | None = None) -> str:
        """ Return the jid we're using in a room.
        """
        self._raise_wrong_pfrom_usage(pfrom)
        if not self.multi_from:
            pfrom = None
        return "%s/%s" % (room_jid, self.our_nicks[pfrom][room_jid])

    def get_jid_property(self, room: JID, nick: str,
                         jid_property: MucRoomItemKeys,
                         pfrom: JID | None = None) -> Any:
        """ Get the property of a nick in a room, such as its 'jid' or 'affiliation'
            If not found, return None.

        :param room: Get the property for this room.
        :param nick: Which nickname information to get.
        :param jid_property: Property to fetch.
        """
        self._raise_wrong_pfrom_usage(pfrom)
        if not self.multi_from:
            pfrom = None
        rooms = self.rooms.get(pfrom, {})
        room_dict = rooms.get(room, {})
        nick_dict = room_dict.get(nick, {})
        prop = nick_dict.get(jid_property)
        return prop or None

    def get_roster(self, room: JID, pfrom: JID | None = None) -> list[str]:
        """ Get the list of nicks in a room.

        :param room: Room to list nicks from.
        """
        self._raise_wrong_pfrom_usage(pfrom)
        if not self.multi_from:
            pfrom = None
        rooms = self.rooms.get(pfrom, {})
        if room not in rooms:
            raise ValueError("Room %s is not joined" % room)
        return list(rooms[room].keys())

    def get_users_by_affiliation(self, room: JidStr, affiliation='member', *, ifrom: JidStr | None = None):
        # Preserve old API
        if affiliation not in AFFILIATIONS:
            raise ValueError("Affiliation %s does not exist" % affiliation)
        return self.get_affiliation_list(room, affiliation, ifrom=ifrom)

    # Aliases muc→groupchat
    join_groupchat = join_muc
    join_groupchat_wait = join_muc_wait



# Slixmpp: The Slick XMPP Library
# Copyright (C) 2025 Mathieu Pasquet
# This file is part of Slixmpp.
# See the file LICENSE for copying permission
import logging
from asyncio import Event, Task, FIRST_COMPLETED, wait, CancelledError
from enum import Enum

from slixmpp.stanza import Message, Presence
from slixmpp.exceptions import IqError, IqTimeout
from slixmpp.jid import JID
from slixmpp.plugins import BasePlugin
from slixmpp.types import ErrorConditions


log = logging.getLogger(__name__)


class PingStatus(Enum):
    """
    Enum representing the status of a ping.

    - `UNTRIED`: the ping was not sent yet
    - `JOINED`: we were considered still in the MUC at the last ping
    - `JOINED_NICKCHANGED`: we were joined but another client just changed
      the nick.
    - `DISCONNECTED`: we got disconnected
    - `TIMEOUT`: slixmpp was unable to reach the server
    """
    UNTRIED = 0
    JOINED = 1
    JOINED_NICKCHANGED = 1
    DISCONNECTED = 2
    TIMEOUT = 3


class PingTask:
    """
    Class representing a current muc self-ping task.
    """
    _event: Event
    _current_task: Task
    timeout: float
    interval: float
    _plugin: 'XEP_0410'

    def __init__(self, muc_resource: JID, orig_jid: JID,
                 plugin: 'XEP_0410', interval: float, timeout: float) -> None:
        self._event = Event()
        self._plugin = plugin
        self.interval = interval
        self.timeout = timeout
        self._current_task = plugin.xmpp.loop.create_task(self.run(
            muc_resource,
            orig_jid,
        ))

    def reset_timer(self) -> None:
        """
        Triggers the internal event to reset the internal timer.
        """
        self._event.set()

    def cancel(self) -> None:
        """Cancel the task before dying"""
        self._current_task.cancel()

    async def run(self, muc_resource: JID, orig_jid: JID) -> None:
        """
        Loop that goes on forever and sends pings at the desired interval.
        """
        while True:
            # Wait on the event or timeout, whichever comes first
            try:
                done, pending = await wait(
                    [self._plugin.xmpp.loop.create_task(self._event.wait())],
                    return_when=FIRST_COMPLETED,
                    timeout=self.interval,
                )
                # If the event is set, then the timer was reset and we clear it
                # before going back to waiting
                if self._event.is_set():
                    self._event.clear()
                    continue
                result = await self._plugin.send_self_ping(
                    muc_resource,
                    orig_jid,
                    timeout=self.timeout,
                )
                key = (muc_resource, orig_jid)
                self._plugin._update_ping_results(key, result)
            except CancelledError:
                return
            except:
                log.exception("Error while trying to ping a muc resource")
                return


class XEP_0410(BasePlugin):

    """
    XEP-0410: MUC Self-Ping (Schrödinger's Chat)

    This plugin provides features for clients (or non-MUC components) to
    enable or disable self-ping.

    Configuration options:

    - `ping_interval`: the number of seconds between the last activity and a ping.

    Ping status is represented using the :class:`~slixmpp.plugins.XEP_0410.PingStatus` class.

    Users of this plugin can either schedule the pings manually and send them
    using `send_self_ping`, or they can use the provided scheduler with
    `enable_self_ping` and `disable_self_ping`.

    Doing so requires calling `update_nick` if your nickname in a MUC changes,
    because the plugin is not aware of this. The scheduled ping will trigger
    a `muc_ping_changed` event in case the ping status changes, which should
    be handled properly.
    """

    name = 'xep_0410'
    description = 'XEP-0410: MUC Self-Ping (Schrödinger\'s Chat)'
    dependencies = {'xep_0045', 'xep_0199'}
    default_config = {
        "ping_interval": 900,
        "ping_timeout": 30,
    }
    ping_interval: int | float
    # Cache of the last bound JID, to be able to recover if we bind to
    # another resource while running
    boundjid: JID | None = None
    # Dictionary mapping a (muc resource, from jid) to an asyncio task in
    # the process of being executed
    ping_timers: dict[tuple[JID, JID], PingTask]
    # Cache of the latest ping results
    last_ping_results: dict[tuple[JID, JID], PingStatus]

    def plugin_init(self):
        self.ping_timers = dict()
        self.boundjid = None
        self.last_ping_results = dict()
        self.xmpp.add_event_handler(
            'groupchat_message',
            self._on_muc_activity
        )
        self.xmpp.add_event_handler(
            'groupchat_presence',
            self._on_muc_activity,
        )

    def plugin_end(self):
        self.ping_timers = dict()
        self.last_ping_results = dict()

    def session_bind(self, jid: JID):
        self.__update_boundjid(jid)

    def __update_boundjid(self, jid: JID):
        """
        If the bound JID has changed, we need to update the timers so that
        the ping stanzas have the correct from value.

        This is a bit painful.
        """
        new_boundjid = jid
        if self.boundjid and self.boundjid != new_boundjid:
            self.ping_results = dict()
            for key, timer in list(self.ping_timers.items()):
                if key[1] == self.boundjid:
                    del self.ping_timers[key]
                    timer.cancel()
                    key = (key[0], new_boundjid)
                    self.ping_timers[key] = PingTask(
                        key[0],
                        key[1],
                        self,
                        timeout=timer.timeout,
                        interval=timer.interval,
                    )
        self.boundjid = jid

    def update_nick(self, previous_jid: JID, new_jid: JID) -> None:
        """
        Update the self-ping targets when a nickname changes.

        :param previous_jid: full JID of the previous MUC resource.
        :param new_jid: full JID of the new MUC resource.
        """
        for key, timer in list(self.ping_timers.items()):
            if key[0] == previous_jid:
                del self.ping_timers[key]
                timer.cancel()
                key = (new_jid, key[1])
                self.ping_timers[key] = PingTask(
                    key[0],
                    key[1],
                    self,
                    timeout=timer.timeout,
                    interval=timer.interval,
                )

    def enable_self_ping(self, muc_resource: JID,
                         orig_jid: JID | None = None,
                         interval: float | None = None,
                         timeout: float | None = None) -> None:
        """
        Enable client self-ping.
        The given MUC resource will be pinged periodically if the MUC is inactive,
        and an event will be generated in case of changes.

        :param muc_resource: The MUC resource to send the ping to.
        :param orig_jid: The "from" resource to send the ping from (for components).
        """
        if orig_jid is None:
            orig_jid = self.xmpp.boundjid
        if timeout is None:
            timeout = self.ping_timeout
        if interval is None:
            interval = self.ping_interval
        key = (muc_resource, orig_jid)
        if key not in self.ping_timers:
            self.ping_timers[key] = PingTask(
                muc_resource=muc_resource,
                orig_jid=orig_jid,
                plugin=self,
                timeout=timeout,
                interval=interval,
            )

    def disable_self_ping(self, muc_resource: JID,
                          orig_jid: JID | None = None) -> None:
        """
        Disable client self-ping. Cancels the scheduled pings for the given
        MUC resource.

        :param muc_resource: The MUC resource to send the ping to.
        :param orig_jid: The "from" resource the ping was sent from (for components).
        """
        if orig_jid is None:
            orig_jid = self.xmpp.boundjid
        key = (muc_resource, orig_jid)
        if key in self.ping_timers:
            self.ping_timers[key].cancel()
            del self.ping_timers[key]
        if key in self.last_ping_results:
            del self.last_ping_results[key]

    def get_ping_status(self, muc_resource: JID,
                        orig_jid: JID | None = None) -> PingStatus:
        """
        Return the last pinged status for a specific muc resource.

        :param muc_resource: The MUC resource to send the ping to.
        :param orig_jid: The "from" resource the ping was sent from (for components).
        :return: The status, or None if no ping has been sent yet.
        """
        if orig_jid is None:
            orig_jid = self.xmpp.boundjid
        key = (muc_resource, orig_jid)
        return self.last_ping_results.get(key, PingStatus.UNTRIED)

    def _on_muc_activity(self, event: Presence | Message):
        """Handle both messages and presences from mucs to see if we need to
        reset the timer"""
        if event['type'] == 'error':
            return
        key = (event['from'], event['to'])
        if key in self.ping_timers:
            self.ping_timers[key].reset_timer()

    def _handle_condition(self, condition: ErrorConditions) -> PingStatus:
        """
        Interpret the error conditions as defined in XEP-0410.
        """
        if condition in ('service-unavailable', 'feature-not-implemented'):
            return PingStatus.JOINED
        elif condition == 'item-not-found':
            return PingStatus.JOINED_NICKCHANGED
        elif condition in ('remote-server-not-found',
                           'remote-server-timeout'):
            return PingStatus.TIMEOUT
        else:
            return PingStatus.DISCONNECTED

    async def send_self_ping(self, muc_resource: JID,
                             orig_jid: JID | None = None,
                             timeout: float | None = None) -> PingStatus:
        """
        Send a single self-ping to a MUC, and return the result.

        :param muc_resource: The MUC resource to send the ping to.
        :param orig_jid: The "from" resource the ping was sent from (for components).
        :return: The ping status
        """
        if orig_jid is None:
            orig_jid = self.xmpp.boundjid
        result = PingStatus.UNTRIED
        try:
            await self.xmpp.plugin['xep_0199'].send_ping(
                muc_resource,
                ifrom=orig_jid,
                timeout=timeout,
            )
            result = PingStatus.JOINED
        except IqTimeout:
            result = PingStatus.TIMEOUT
        except IqError as exc:
            result = self._handle_condition(exc.condition)
        return result

    def _update_ping_results(self, key: tuple[JID, JID], result: PingStatus) -> None:
        """
        Internal use only: used to update the ping results dict from the timer.
        """
        previous = self.last_ping_results.get(key, PingStatus.UNTRIED)
        self.last_ping_results[key] = result
        if result != previous:
            self.xmpp.event('muc_ping_changed', {
                "key": key,
                "previous": previous,
                "result": result
            })




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2010 Nathanael C. Fritz, Lance J.T. Stout
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
import asyncio
import logging

from asyncio import Future
from typing import (
    Callable,
)

from slixmpp import JID
from slixmpp.stanza import Iq
from slixmpp.types import OptJid
from slixmpp.plugins import BasePlugin
from slixmpp.xmlstream.handler import CoroutineCallback
from slixmpp.xmlstream.matcher import StanzaPath
from slixmpp.xmlstream import register_stanza_plugin
from slixmpp.plugins.xep_0030 import stanza, DiscoInfo, DiscoItems
from slixmpp.plugins.xep_0030 import StaticDisco


log = logging.getLogger(__name__)


class XEP_0030(BasePlugin):

    """
    XEP-0030: Service Discovery

    Service discovery in XMPP allows entities to discover information about
    other agents in the network, such as the feature sets supported by a
    client, or signposts to other, related entities.

    Also see <http://www.xmpp.org/extensions/xep-0030.html>.

    The XEP-0030 plugin works using a hierarchy of dynamic
    node handlers, ranging from global handlers to specific
    JID+node handlers. The default set of handlers operate
    in a static manner, storing disco information in memory.
    However, custom handlers may use any available backend
    storage mechanism desired, such as SQLite or Redis.

    Node handler hierarchy:

    ====== ======= ============================
    JID     Node    Level
    ====== ======= ============================
    None    None    Global
    Given   None    All nodes for the JID
    None    Given   Node on self.xmpp.boundjid
    Given   Given   A single node
    ====== ======= ============================

    Adding information for a given node without specifying the JID will
    use the bound JID and therefore must be done after the bind happens.

    Stream Handlers:

    ::

        Disco Info  -- Any Iq stanze that includes a query with the
                       namespace http://jabber.org/protocol/disco#info.
        Disco Items -- Any Iq stanze that includes a query with the
                       namespace http://jabber.org/protocol/disco#items.

    Events:

    - :term:`disco_info`         -- Received a disco#info Iq query result.
    - :term:`disco_items`        -- Received a disco#items Iq query result.

    Attributes:

    :var static: Object containing the default set of
                 static node handlers.
    """

    name = 'xep_0030'
    description = 'XEP-0030: Service Discovery'
    dependencies = set()
    stanza = stanza
    default_config = {
        'use_cache': True,
        'wrap_results': False
    }
    static: StaticDisco

    def plugin_init(self):
        """
        Start the XEP-0030 plugin.
        """
        self.xmpp.register_handler(CoroutineCallback(
            'Disco Info',
            StanzaPath('iq/disco_info'),
            self._handle_disco_info
        ))

        self.xmpp.register_handler(CoroutineCallback(
            'Disco Items',
            StanzaPath('iq/disco_items'),
            self._handle_disco_items
        ))

        register_stanza_plugin(Iq, DiscoInfo)
        register_stanza_plugin(Iq, DiscoItems)

        self.static = StaticDisco(self.xmpp, self)

        self._disco_ops = [
                'get_info', 'set_info', 'set_identities', 'set_features',
                'get_items', 'set_items', 'del_items', 'add_identity',
                'del_identity', 'add_feature', 'del_feature', 'add_item',
                'del_item', 'del_identities', 'del_features', 'cache_info',
                'get_cached_info', 'supports', 'has_identity']

        for op in self._disco_ops:
            self.api.register(getattr(self.static, op), op, default=True)

        self.domain_infos = {}

    def session_bind(self, jid):
        self.add_feature('http://jabber.org/protocol/disco#info')

    def plugin_end(self):
        self.del_feature('http://jabber.org/protocol/disco#info')

    def _add_disco_op(self, op, default_handler):
        self.api.register(default_handler, op)
        self.api.register_default(default_handler, op)

    def set_node_handler(self, htype: str, jid: OptJid = None,
                         node: str | None = None,
                         handler: Callable | None = None):
        """
        Add a node handler for the given hierarchy level and
        handler type.

        Node handlers are ordered in a hierarchy where the
        most specific handler is executed. Thus, a fallback,
        global handler can be used for the majority of cases
        with a few node specific handler that override the
        global behavior.

        Node handler hierarchy:

        ====== ======= ============================
        JID     Node    Level
        ====== ======= ============================
        None    None    Global
        Given   None    All nodes for the JID
        None    Given   Node on self.xmpp.boundjid
        Given   Given   A single node
        ====== ======= ============================

        Handler types:

        ::

            get_info
            get_items
            set_identities
            set_features
            set_items
            del_items
            del_identities
            del_identity
            del_feature
            del_features
            del_item
            add_identity
            add_feature
            add_item

        :param htype: The operation provided by the handler.
        :param jid: The JID the handler applies to. May be narrowed
                    further if a node is given.
        :param node: The particular node the handler is for. If no JID
                     is given, then the self.xmpp.boundjid.full is
                     assumed.
        :param handler: The handler function to use.
        """
        self.api.register(handler, htype, jid, node)

    def del_node_handler(self, htype: str, jid: OptJid, node: str | None):
        """
        Remove a handler type for a JID and node combination.

        The next handler in the hierarchy will be used if one
        exists. If removing the global handler, make sure that
        other handlers exist to process existing nodes.

        Node handler hierarchy:

        ====== ======= ============================
        JID     Node    Level
        ====== ======= ============================
        None    None    Global
        Given   None    All nodes for the JID
        None    Given   Node on self.xmpp.boundjid
        Given   Given   A single node
        ====== ======= ============================

        :param htype: The type of handler to remove.
        :param jid: The JID from which to remove the handler.
        :param node: The node from which to remove the handler.
        """
        self.api.unregister(htype, jid, node)

    def restore_defaults(self, jid: OptJid = None, node: str | None = None,
                         handlers: list[Callable] | None = None):
        """
        Change all or some of a node's handlers to the default
        handlers. Useful for manually overriding the contents
        of a node that would otherwise be handled by a JID level
        or global level dynamic handler.

        The default is to use the built-in static handlers, but that
        may be changed by modifying self.default_handlers.

        :param jid: The JID owning the node to modify.
        :param node: The node to change to using static handlers.
        :param handlers: Optional list of handlers to change to the
                         default version. If provided, only these
                         handlers will be changed. Otherwise, all
                         handlers will use the default version.
        """
        if handlers is None:
            handlers = self._disco_ops
        for op in handlers:
            self.api.restore_default(op, jid, node)

    def supports(self, jid: OptJid = None, node: str | None = None,
                 feature: str | None = None, local: bool = False,
                 cached: bool = True, ifrom: OptJid = None) -> Future:
        """
        Check if a JID supports a given feature.

        .. versionchanged:: 1.8.0
            This function now returns a Future.

        :param jid: Request info from this JID.
        :param node: The particular node to query.
        :param feature: The name of the feature to check.
        :param local: If true, then the query is for a JID/node
                      combination handled by this Slixmpp instance and
                      no stanzas need to be sent.
                      Otherwise, a disco stanza must be sent to the
                      remove JID to retrieve the info.
        :param cached: If true, then look for the disco info data from
                       the local cache system. If no results are found,
                       send the query as usual. The self.use_cache
                       setting must be set to true for this option to
                       be useful. If set to false, then the cache will
                       be skipped, even if a result has already been
                       cached. Defaults to false.

        :returns True: The feature is supported
        :returns False: The feature is not listed as supported
        :returns None: Nothing could be found due to a timeout
        """
        data = {'feature': feature,
                'local': local,
                'cached': cached}
        return self.api['supports'](jid, node, ifrom, data)

    def has_identity(self, jid: OptJid = None, node: str | None = None,
                     category: str | None = None,
                     itype: str | None = None, lang: str | None = None,
                     local: bool = False, cached: bool = True,
                     ifrom: OptJid = None) -> Future:
        """
        Check if a JID provides a given identity.

        .. versionchanged:: 1.8.0
            This function now returns a Future.


        :param jid: Request info from this JID.
        :param node: The particular node to query.
        :param category: The category of the identity to check.
        :param itype: The type of the identity to check.
        :param lang: The language of the identity to check.
        :param local: If true, then the query is for a JID/node
                        combination handled by this Slixmpp instance and
                        no stanzas need to be sent.
                        Otherwise, a disco stanza must be sent to the
                        remove JID to retrieve the info.
        :param cached: If true, then look for the disco info data from
                        the local cache system. If no results are found,
                        send the query as usual. The self.use_cache
                        setting must be set to true for this option to
                        be useful. If set to false, then the cache will
                        be skipped, even if a result has already been
                        cached. Defaults to false.

        :returns True: The identity is provided
        :returns False: The identity is not listed
        :returns None: Nothing could be found due to a timeout
        """
        data = {'category': category,
                'itype': itype,
                'lang': lang,
                'local': local,
                'cached': cached}
        return self.api['has_identity'](jid, node, ifrom, data)

    async def get_info_from_domain(self, domain=None, timeout=None,
                                   cached=True, callback=None, **iqkwargs):
        """Fetch disco#info of specified domain and one disco#items level below
        """

        if domain is None:
            domain = self.xmpp.boundjid.domain

        if not cached or domain not in self.domain_infos:
            infos = [asyncio.create_task(self.get_info(
                domain, timeout=timeout, **iqkwargs))]
            iq_items = await self.get_items(
                domain, timeout=timeout, **iqkwargs)
            items = iq_items['disco_items']['items']
            infos += [
                asyncio.create_task(self.get_info(item[0], timeout=timeout, **iqkwargs))
                for item in items]
            info_futures, _ = await asyncio.wait(
                infos,
                timeout=timeout,
            )

            self.domain_infos[domain] = [
                future.result()
                for future in info_futures if not future.exception()
            ]

        results = self.domain_infos[domain]

        if callback is not None:
            callback(results)
        return results

    async def get_info(self, jid: OptJid = None, node: str | None = None,
                       local: bool | None = None,
                       cached: bool | None = None, **kwargs) -> Iq:
        """
        Retrieve the disco#info results from a given JID/node combination.

        Info may be retrieved from both local resources and remote agents;
        the local parameter indicates if the information should be gathered
        by executing the local node handlers, or if a disco#info stanza
        must be generated and sent.

        If requesting items from a local JID/node, then only a DiscoInfo
        stanza will be returned. Otherwise, an Iq stanza will be returned.

        .. versionchanged:: 1.8.0
            This function is now a coroutine.

        :param jid: Request info from this JID.
        :param node: The particular node to query.
        :param local: If true, then the query is for a JID/node
                      combination handled by this Slixmpp instance and
                      no stanzas need to be sent.
                      Otherwise, a disco stanza must be sent to the
                      remote JID to retrieve the info.
        :param cached: If true, then look for the disco info data from
                       the local cache system. If no results are found,
                       send the query as usual. The self.use_cache
                       setting must be set to true for this option to
                       be useful. If set to false, then the cache will
                       be skipped, even if a result has already been
                       cached. Defaults to false.
        """
        if local is None:
            if jid is not None and not isinstance(jid, JID):
                jid = JID(jid)
                if self.xmpp.is_component:
                    if jid.domain == self.xmpp.boundjid.domain:
                        local = True
                else:
                    if str(jid) == str(self.xmpp.boundjid):
                        local = True
                jid = jid.full
            elif jid in (None, ''):
                local = True

        ifrom = kwargs.pop('ifrom', None)
        if self.xmpp.is_component and ifrom is None:
            ifrom = self.xmpp.boundjid
        if local:
            log.debug("Looking up local disco#info data "
                      "for %s, node %s.", jid, node)
            info = await self.api['get_info'](
                jid, node, ifrom,
                kwargs
            )
            info = self._fix_default_info(info)
            return self._wrap(ifrom, jid, info)

        if cached:
            log.debug("Looking up cached disco#info data "
                      "for %s, node %s.", jid, node)
            info = await self.api['get_cached_info'](
                jid, node, ifrom,
                kwargs
            )
            if info is not None:
                return self._wrap(ifrom, jid, info)

        iq = self.xmpp.Iq()
        # Check dfrom parameter for backwards compatibility
        iq['from'] = ifrom or kwargs.get('dfrom', '')
        iq['to'] = jid
        iq['type'] = 'get'
        iq['disco_info']['node'] = node if node else ''
        return await iq.send(**kwargs)

    def set_info(self, jid: OptJid = None, node: str | None = None,
                 info: Iq | DiscoInfo | None = None) -> Future:
        """
        Set the disco#info data for a JID/node based on an existing
        disco#info stanza.

        .. versionchanged:: 1.8.0
            This function now returns a Future.

        """
        if isinstance(info, Iq):
            info = info['disco_info']
        return self.api['set_info'](jid, node, None, info)

    async def get_items(self, jid: OptJid = None, node: str | None = None,
                        local: bool = False, iterator: bool = False,
                        ifrom: OptJid = None, **kwargs) -> Iq:
        """
        Retrieve the disco#items results from a given JID/node combination.

        Items may be retrieved from both local resources and remote agents;
        the local parameter indicates if the items should be gathered by
        executing the local node handlers, or if a disco#items stanza must
        be generated and sent.

        If requesting items from a local JID/node, then only a DiscoItems
        stanza will be returned. Otherwise, an Iq stanza will be returned.

        .. versionchanged:: 1.8.0
            This function is now a coroutine.

        :param jid: Request info from this JID.
        :param node: The particular node to query.
        :param local: If true, then the query is for a JID/node
                      combination handled by this Slixmpp instance and
                      no stanzas need to be sent.
                      Otherwise, a disco stanza must be sent to the
                      remove JID to retrieve the items.
        :param iterator: If True, return a result set iterator using
                         the XEP-0059 plugin, if the plugin is loaded.
                         Otherwise the parameter is ignored.
        """
        if ifrom is None and self.xmpp.is_component:
            ifrom = self.xmpp.boundjid.bare

        if local or local is None and jid is None:
            items = await self.api['get_items'](jid, node, ifrom, kwargs)
            return self._wrap(ifrom, jid, items)

        iq = self.xmpp.Iq()
        # Check dfrom parameter for backwards compatibility
        iq['from'] = ifrom or kwargs.get('dfrom', '')
        iq['to'] = jid
        iq['type'] = 'get'
        iq['disco_items']['node'] = node if node else ''
        if iterator and self.xmpp['xep_0059']:
            return self.xmpp['xep_0059'].iterate(iq, 'disco_items')
        else:
            return await iq.send(**kwargs)

    def set_items(self, jid: OptJid = None, node: str | None = None,
                  **kwargs) -> Future:
        """
        Set or replace all items for the specified JID/node combination.

        The given items must be in a list or set where each item is a
        tuple of the form: (jid, node, name).


        .. versionchanged:: 1.8.0
            This function now returns a Future.

        :param jid: The JID to modify.
        :param node: Optional node to modify.
        :param items: A series of items in tuple format.
        """
        return self.api['set_items'](jid, node, None, kwargs)

    def del_items(self, jid: OptJid = None, node: str | None = None,
                  **kwargs) -> Future:
        """
        Remove all items from the given JID/node combination.

        .. versionchanged:: 1.8.0
            This function now returns a Future.

        Arguments:
        :param jid: The JID to modify.
        :param node: Optional node to modify.
        """
        return self.api['del_items'](jid, node, None, kwargs)

    def add_item(self, jid: str = '', name: str = '',
                 node: str | None = None, subnode: str = '',
                 ijid: OptJid = None) -> Future:
        """
        Add a new item element to the given JID/node combination.

        Each item is required to have a JID, but may also specify
        a node value to reference non-addressable entities.

        .. versionchanged:: 1.8.0
            This function now returns a Future.

        :param jid: The JID for the item.
        :param name: Optional name for the item.
        :param node: The node to modify.
        :param subnode: Optional node for the item.
        :param ijid: The JID to modify.
        """
        if not jid:
            jid = self.xmpp.boundjid.full
        kwargs = {'ijid': jid,
                  'name': name,
                  'inode': subnode}
        return self.api['add_item'](ijid, node, None, kwargs)

    def del_item(self, jid: OptJid = None, node: str | None = None,
                 **kwargs) -> Future:
        """
        Remove a single item from the given JID/node combination.

        :param jid: The JID to modify.
        :param node: The node to modify.
        :param ijid: The item's JID.
        :param inode: The item's node.
        """
        return self.api['del_item'](jid, node, None, kwargs)

    def add_identity(self, category: str = '', itype: str = '', name: str = '',
                     node: str | None = None, jid: OptJid = None,
                     lang: str | None = None) -> Future:
        """
        Add a new identity to the given JID/node combination.

        Each identity must be unique in terms of all four identity
        components: category, type, name, and language.

        Multiple, identical category/type pairs are allowed only
        if the xml:lang values are different. Likewise, multiple
        category/type/xml:lang pairs are allowed so long as the
        names are different. A category and type is always required.

        .. versionchanged:: 1.8.0
            This function now returns a Future.

        :param category: The identity's category.
        :param itype: The identity's type.
        :param name: Optional name for the identity.
        :param lang: Optional two-letter language code.
        :param node: The node to modify.
        :param jid: The JID to modify.
        """
        kwargs = {'category': category,
                  'itype': itype,
                  'name': name,
                  'lang': lang}
        return self.api['add_identity'](jid, node, None, kwargs)

    def add_feature(self, feature: str, node: str | None = None,
                    jid: OptJid = None) -> Future:
        """
        Add a feature to a JID/node combination.

        .. versionchanged:: 1.8.0
            This function now returns a Future.

        :param feature: The namespace of the supported feature.
        :param node: The node to modify.
        :param jid: The JID to modify.
        """
        kwargs = {'feature': feature}
        return self.api['add_feature'](jid, node, None, kwargs)

    def del_identity(self, jid: OptJid = None,
                     node: str | None = None, **kwargs) -> Future:
        """
        Remove an identity from the given JID/node combination.

        .. versionchanged:: 1.8.0
            This function now returns a Future.

        :param jid: The JID to modify.
        :param node: The node to modify.
        :param category: The identity's category.
        :param itype: The identity's type value.
        :param name: Optional, human readable name for the identity.
        :param lang: Optional, the identity's xml:lang value.
        """
        return self.api['del_identity'](jid, node, None, kwargs)

    def del_feature(self, jid: OptJid = None, node: str | None = None,
                    **kwargs) -> Future:
        """
        Remove a feature from a given JID/node combination.

        .. versionchanged:: 1.8.0
            This function now returns a Future.

        :param jid: The JID to modify.
        :param node: The node to modify.
        :param feature: The feature's namespace.
        """
        return self.api['del_feature'](jid, node, None, kwargs)

    def set_identities(self, jid: OptJid = None, node: str | None = None,
                       **kwargs) -> Future:
        """
        Add or replace all identities for the given JID/node combination.

        The identities must be in a set where each identity is a tuple
        of the form: (category, type, lang, name)

        .. versionchanged:: 1.8.0
            This function now returns a Future.

        :param jid: The JID to modify.
        :param node: The node to modify.
        :param identities: A set of identities in tuple form.
        :param lang: Optional, xml:lang value.
        """
        return self.api['set_identities'](jid, node, None, kwargs)

    def del_identities(self, jid: OptJid = None, node: str | None = None,
                       **kwargs) -> Future:
        """
        Remove all identities for a JID/node combination.

        If a language is specified, only identities using that
        language will be removed.

        .. versionchanged:: 1.8.0
            This function now returns a Future.

        :param jid: The JID to modify.
        :param node: The node to modify.
        :param lang: Optional. If given, only remove identities
                    using this xml:lang value.
        """
        return self.api['del_identities'](jid, node, None, kwargs)

    def set_features(self, jid: OptJid = None, node: str | None = None,
                     **kwargs) -> Future:
        """
        Add or replace the set of supported features
        for a JID/node combination.

        .. versionchanged:: 1.8.0
            This function now returns a Future.

        :param jid: The JID to modify.
        :param node: The node to modify.
        :param features: The new set of supported features.
        """
        return self.api['set_features'](jid, node, None, kwargs)

    def del_features(self, jid: OptJid = None, node: str | None = None,
                     **kwargs) -> Future:
        """
        Remove all features from a JID/node combination.

        .. versionchanged:: 1.8.0
            This function now returns a Future.

        :param jid: The JID to modify.
        :param node: The node to modify.
        """
        return self.api['del_features'](jid, node, None, kwargs)

    async def _run_node_handler(self, htype, jid, node: str | None = None,
                                ifrom: OptJid = None, data=None):
        """
        Execute the most specific node handler for the given
        JID/node combination.

        :param htype: The handler type to execute.
        :param jid: The JID requested.
        :param node: The node requested.
        :param data: Optional, custom data to pass to the handler.
        """
        if not data:
            data = {}

        return await self.api[htype](jid, node, ifrom, data)

    async def _handle_disco_info(self, iq: Iq):
        """
        Process an incoming disco#info stanza. If it is a get
        request, find and return the appropriate identities
        and features. If it is an info result, fire the
        disco_info event.

        :param iq: The incoming disco#items stanza.
        """
        if iq['type'] == 'get':
            log.debug("Received disco info query from "
                      "<%s> to <%s>.", iq['from'], iq['to'])
            info = await self.api['get_info'](iq['to'],
                                              iq['disco_info']['node'],
                                              iq['from'],
                                              iq)
            if isinstance(info, Iq):
                info['id'] = iq['id']
                info.send()
            else:
                node = iq['disco_info']['node']
                iq = iq.reply()
                if info:
                    info = self._fix_default_info(info)
                    info['node'] = node
                    iq.set_payload(info.xml)
                iq.send()
        elif iq['type'] == 'result':
            log.debug("Received disco info result from "
                      "<%s> to <%s>.", iq['from'], iq['to'])
            if self.use_cache:
                log.debug("Caching disco info result from "
                          "<%s> to <%s>.", iq['from'], iq['to'])
                if self.xmpp.is_component:
                    ito = iq['to'].full
                else:
                    ito = None
                await self.api['cache_info'](iq['from'],
                                             iq['disco_info']['node'],
                                             ito,
                                             iq)
            self.xmpp.event('disco_info', iq)

    async def _handle_disco_items(self, iq: Iq):
        """
        Process an incoming disco#items stanza. If it is a get
        request, find and return the appropriate items. If it
        is an items result, fire the disco_items event.

        :param iq: The incoming disco#items stanza.
        """
        if iq['type'] == 'get':
            log.debug("Received disco items query from "
                      "<%s> to <%s>.", iq['from'], iq['to'])
            items = await self.api['get_items'](iq['to'],
                                                iq['disco_items']['node'],
                                                iq['from'],
                                                iq)
            if isinstance(items, Iq):
                items.send()
            else:
                iq = iq.reply()
                if items:
                    iq.set_payload(items.xml)
                iq.send()
        elif iq['type'] == 'result':
            log.debug("Received disco items result from "
                      "%s to %s.", iq['from'], iq['to'])
            self.xmpp.event('disco_items', iq)

    def _fix_default_info(self, info: DiscoInfo):
        """
        Disco#info results for a JID are required to include at least
        one identity and feature. As a default, if no other identity is
        provided, Slixmpp will use either the generic component or the
        bot client identity. A the standard disco#info feature will also be
        added if no features are provided.

        :param info: The disco#info quest (not the full Iq stanza) to modify.
        """
        result = info
        if isinstance(info, Iq):
            info = info['disco_info']
        if not info['node']:
            if not info['identities']:
                if self.xmpp.is_component:
                    log.debug("No identity found for this entity. "
                              "Using default component identity.")
                    info.add_identity('component', 'generic')
                else:
                    log.debug("No identity found for this entity. "
                              "Using default client identity.")
                    info.add_identity('client', 'bot')
            if not info['features']:
                log.debug("No features found for this entity. "
                          "Using default disco#info feature.")
                info.add_feature(info.namespace)
        return result

    def _wrap(self, ito: OptJid, ifrom: OptJid, payload, force=False) -> Iq:
        """
        Ensure that results are wrapped in an Iq stanza
        if self.wrap_results has been set to True.

        :param ito: The JID to use as the 'to' value
        :param ifrom: The JID to use as the 'from' value
        :param payload: The disco data to wrap
        :param force: Force wrapping, regardless of self.wrap_results
        """
        if (force or self.wrap_results) and not isinstance(payload, Iq):
            iq = self.xmpp.Iq()
            # Since we're simulating a result, we have to treat
            # the 'from' and 'to' values opposite the normal way.
            iq['to'] = self.xmpp.boundjid if ito is None else ito
            iq['from'] = self.xmpp.boundjid if ifrom is None else ifrom
            iq['type'] = 'result'
            iq.append(payload)
            return iq
        return payload




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2012 Nathanael C. Fritz, Lance J.T. Stout
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
import hmac
import hashlib
import urllib
import base64

from slixmpp.xmlstream import ElementBase


class OAuth(ElementBase):

    name = 'oauth'
    namespace = 'urn:xmpp:oauth:0'
    plugin_attrib = 'oauth'
    interfaces = {'oauth_consumer_key', 'oauth_nonce', 'oauth_signature',
                  'oauth_signature_method', 'oauth_timestamp', 'oauth_token',
                  'oauth_version'}
    sub_interfaces = interfaces

    def generate_signature(self, stanza, sfrom, sto, consumer_secret,
                                 token_secret, method='HMAC-SHA1'):
        self['oauth_signature_method'] = method

        request = urllib.parse.quote('%s&%s' % (sfrom, sto), '')
        parameters = urllib.parse.quote('&'.join([
            'oauth_consumer_key=%s' % self['oauth_consumer_key'],
            'oauth_nonce=%s' % self['oauth_nonce'],
            'oauth_signature_method=%s' % self['oauth_signature_method'],
            'oauth_timestamp=%s' % self['oauth_timestamp'],
            'oauth_token=%s' % self['oauth_token'],
            'oauth_version=%s' % self['oauth_version']
        ]), '')

        sigbase = '%s&%s&%s' % (stanza, request, parameters)

        consumer_secret = urllib.parse.quote(consumer_secret, '')
        token_secret = urllib.parse.quote(token_secret, '')
        key = '%s&%s' % (consumer_secret, token_secret)

        if method == 'HMAC-SHA1':
            sig = base64.b64encode(hmac.new(key, sigbase, hashlib.sha1).digest())
        elif method == 'PLAINTEXT':
            sig = key
        else:
            raise ValueError('Unknown signature method')

        self['oauth_signature'] = sig
        return sig

    def verify_signature(self, stanza, sfrom, sto, consumer_secret,
                               token_secret):
        method = self['oauth_signature_method']

        request = urllib.parse.quote('%s&%s' % (sfrom, sto), '')
        parameters = urllib.parse.quote('&'.join([
            'oauth_consumer_key=%s' % self['oauth_consumer_key'],
            'oauth_nonce=%s' % self['oauth_nonce'],
            'oauth_signature_method=%s' % self['oauth_signature_method'],
            'oauth_timestamp=%s' % self['oauth_timestamp'],
            'oauth_token=%s' % self['oauth_token'],
            'oauth_version=%s' % self['oauth_version']
        ]), '')

        sigbase = '%s&%s&%s' % (stanza, request, parameters)

        consumer_secret = urllib.parse.quote(consumer_secret, '')
        token_secret = urllib.parse.quote(token_secret, '')
        key = '%s&%s' % (consumer_secret, token_secret)

        if method == 'HMAC-SHA1':
            sig = base64.b64encode(hmac.new(key, sigbase, hashlib.sha1).digest())
        elif method == 'PLAINTEXT':
            sig = key
        else:
            raise ValueError('Unknown signature method')

        return self['oauth_signature'] == sig




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2011  Nathanael C. Fritz
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
import base64

from slixmpp.util import bytes
from slixmpp.xmlstream import StanzaBase


class Response(StanzaBase):

    """
    """

    name = 'response'
    namespace = 'urn:ietf:params:xml:ns:xmpp-sasl'
    interfaces = {'value'}
    plugin_attrib = name

    def setup(self, xml):
        StanzaBase.setup(self, xml)
        self.xml.tag = self.tag_name()

    def get_value(self):
        return base64.b64decode(bytes(self.xml.text))

    def set_value(self, values):
        if values:
            self.xml.text = bytes(base64.b64encode(values)).decode('utf-8')
        else:
            self.xml.text = '='

    def del_value(self):
        self.xml.text = ''



# Slixmpp: The Slick XMPP Library
# Copyright (C) 2010 Nathanael C. Fritz, Lance J.T. Stout
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
from typing import (
    Iterable,
)
from slixmpp.xmlstream import ElementBase, ET

IdentityType = tuple[str, str, str | None, str | None]


class DiscoInfo(ElementBase):

    """
    XMPP allows for users and agents to find the identities and features
    supported by other entities in the XMPP network through service discovery,
    or "disco". In particular, the "disco#info" query type for <iq> stanzas is
    used to request the list of identities and features offered by a JID.

    An identity is a combination of a category and type, such as the 'client'
    category with a type of 'pc' to indicate the agent is a human operated
    client with a GUI, or a category of 'gateway' with a type of 'aim' to
    identify the agent as a gateway for the legacy AIM protocol. See
    `XMPP Registrar Disco Categories`_ for a full list of
    accepted category and type combinations.

    .. _XMPP Registrar Disco Categories: <http://xmpp.org/registrar/disco-categories.html>

    Features are simply a set of the namespaces that identify the supported
    features. For example, a client that supports service discovery will
    include the feature ``http://jabber.org/protocol/disco#info``.

    Since clients and components may operate in several roles at once, identity
    and feature information may be grouped into "nodes". If one were to write
    all of the identities and features used by a client, then node names would
    be like section headings.

    Example disco#info stanza:

    .. code-block:: xml

        <iq type="get">
          <query xmlns="http://jabber.org/protocol/disco#info" />
        </iq>

        <iq type="result">
          <query xmlns="http://jabber.org/protocol/disco#info">
            <identity category="client" type="bot" name="Slixmpp Bot" />
            <feature var="http://jabber.org/protocol/disco#info" />
            <feature var="jabber:x:data" />
            <feature var="urn:xmpp:ping" />
          </query>
        </iq>
    """

    name = 'query'
    namespace = 'http://jabber.org/protocol/disco#info'
    plugin_attrib = 'disco_info'
    #: Stanza interfaces:
    #:
    #: - ``node``: The name of the node to either query or return the info from
    #: - ``identities``: A set of 4-tuples, where each tuple contains the
    #:   category, type, xml:lang and name of an identity
    #: - ``features``: A set of namespaces for features
    #:
    interfaces = {'node', 'features', 'identities'}
    lang_interfaces = {'identities'}

    # Cache identities and features
    _identities: set[tuple[str, str, str | None]]
    _features: set[str]

    def setup(self, xml: ET.ElementTree | None = None):
        """
        Populate the stanza object using an optional XML object.

        Overrides ElementBase.setup

        Caches identity and feature information.

        :param xml: Use an existing XML object for the stanza's values.
        """
        ElementBase.setup(self, xml)

        self._identities = {id[0:3] for id in self['identities']}
        self._features = self['features']

    def add_identity(self, category: str, itype: str,
                     name: str | None = None, lang: str | None = None
                     ) -> bool:
        """
        Add a new identity element. Each identity must be unique
        in terms of all four identity components.

        Multiple, identical category/type pairs are allowed only
        if the xml:lang values are different. Likewise, multiple
        category/type/xml:lang pairs are allowed so long as the names
        are different. In any case, a category and type are required.

        :param category: The general category to which the agent belongs.
        :param itype: A more specific designation with the category.
        :param name: Optional human readable name for this identity.
        :param lang: Optional standard xml:lang value.
        """
        identity = (category, itype, lang)
        if identity not in self._identities:
            self._identities.add(identity)
            id_xml = ET.Element('{%s}identity' % self.namespace)
            id_xml.attrib['category'] = category
            id_xml.attrib['type'] = itype
            if lang:
                id_xml.attrib['{%s}lang' % self.xml_ns] = lang
            if name:
                id_xml.attrib['name'] = name
            self.xml.insert(0, id_xml)
            return True
        return False

    def del_identity(self, category: str, itype: str, name=None,
                     lang: str | None = None) -> bool:
        """
        Remove a given identity.

        :param category: The general category to which the agent belonged.
        :param itype: A more specific designation with the category.
        :param name: Optional human readable name for this identity.
        :param lang: Optional, standard xml:lang value.
        """
        identity = (category, itype, lang)
        if identity in self._identities:
            self._identities.remove(identity)
            for id_xml in self.xml.findall('{%s}identity' % self.namespace):
                id = (id_xml.attrib['category'],
                      id_xml.attrib['type'],
                      id_xml.attrib.get('{%s}lang' % self.xml_ns, None))
                if id == identity:
                    self.xml.remove(id_xml)
                    return True
        return False

    def dict_identities(self, lang: str | None = None) -> list[dict[str, str]]:
        """
        Return the list of all identities, each one as a dict with
        category, type, xml_lang, and name keys.

        :param lang: If there is a need to filter identities by lang.
        """
        ids = self.get_identities(lang=lang, dedupe=True)
        dict_ids = []
        for identity in ids:
            dict_ids.append({
                'category': identity[0],
                'type': identity[1],
                'xml_lang': identity[2],
                'name': identity[3],
            })
        return dict_ids

    def get_identities(self, lang: str | None = None, dedupe: bool = True
                       ) -> Iterable[IdentityType]:
        """
        Return a set of all identities in tuple form as so:

            (category, type, lang, name)

        If a language was specified, only return identities using
        that language.

        :param lang: Optional, standard xml:lang value.
        :param dedupe: If True, de-duplicate identities, otherwise
                       return a list of all identities.
        """
        identities: list[IdentityType] | set[IdentityType]
        if dedupe:
            identities = set()
        else:
            identities = []
        for id_xml in self.xml.findall('{%s}identity' % self.namespace):
            xml_lang = id_xml.attrib.get('{%s}lang' % self.xml_ns, None)
            category = id_xml.attrib.get('category', None)
            type_ = id_xml.attrib.get('type', None)
            name = id_xml.attrib.get('name', None)
            if lang is None or xml_lang == lang:
                id = (category, type_, xml_lang, name)
                if isinstance(identities, set):
                    identities.add(id)
                else:
                    identities.append(id)
        return identities

    def set_identities(self, identities: Iterable[IdentityType],
                       lang: str | None = None):
        """
        Add or replace all identities. The identities must be a in set
        where each identity is a tuple of the form:

            (category, type, lang, name)

        If a language is specified, any identities using that language
        will be removed to be replaced with the given identities.

        .. note::

            An identity's language will not be changed regardless of
            the value of lang.

        :param identities: A set of identities in tuple form.
        :param lang: Optional, standard xml:lang value.
        """
        self.del_identities(lang)
        for identity in identities:
            category, itype, lang, name = identity
            self.add_identity(category, itype, name, lang)

    def del_identities(self, lang: str | None = None):
        """
        Remove all identities. If a language was specified, only
        remove identities using that language.

        :param lang: Optional, standard xml:lang value.
        """
        for id_xml in self.xml.findall('{%s}identity' % self.namespace):
            if lang is None:
                self.xml.remove(id_xml)
            elif id_xml.attrib.get('{%s}lang' % self.xml_ns, None) == lang:
                self._identities.remove((
                    id_xml.attrib['category'],
                    id_xml.attrib['type'],
                    id_xml.attrib.get('{%s}lang' % self.xml_ns, None)))
                self.xml.remove(id_xml)

    def add_feature(self, feature: str) -> bool:
        """
        Add a single, new feature.

        :param feature: The namespace of the supported feature.
        """
        if feature not in self._features:
            self._features.add(feature)
            feature_xml = ET.Element('{%s}feature' % self.namespace)
            feature_xml.attrib['var'] = feature
            self.xml.append(feature_xml)
            return True
        return False

    def del_feature(self, feature: str) -> bool:
        """
        Remove a single feature.

        :param feature: The namespace of the removed feature.
        """
        if feature in self._features:
            self._features.remove(feature)
            for feature_xml in self.xml.findall('{%s}feature' % self.namespace):
                if feature_xml.attrib['var'] == feature:
                    self.xml.remove(feature_xml)
                    return True
        return False

    def get_features(self, dedupe: bool = True) -> Iterable[str]:
        """Return the set of all supported features."""
        features: list[str] | set[str]
        if dedupe:
            features = set()
        else:
            features = []
        for feature_xml in self.xml.findall('{%s}feature' % self.namespace):
            feature = feature_xml.attrib.get('var', None)
            if feature:
                if isinstance(features, set):
                    features.add(feature)
                else:
                    features.append(feature)
        return features

    def set_features(self, features: Iterable[str]):
        """
        Add or replace the set of supported features.

        :param features: The new set of supported features.
        """
        self.del_features()
        for feature in features:
            self.add_feature(feature)

    def del_features(self):
        """Remove all features."""
        self._features = set()
        for feature_xml in self.xml.findall('{%s}feature' % self.namespace):
            self.xml.remove(feature_xml)




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2012 Nathanael C. Fritz, Lance J.T. Stout
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
from base64 import b64encode, b64decode

from slixmpp.util import bytes
from slixmpp.xmlstream import ElementBase, register_stanza_plugin


class Data(ElementBase):
    name = 'data'
    namespace = 'urn:xmpp:avatar:data'
    plugin_attrib = 'avatar_data'
    interfaces = {'value'}

    def get_value(self):
        if self.xml.text:
            return b64decode(bytes(self.xml.text))
        return b''

    def set_value(self, value):
        if value:
            self.xml.text = b64encode(bytes(value)).decode()
        else:
            self.xml.text = ''

    def del_value(self):
        self.xml.text = ''


class MetaData(ElementBase):
    name = 'metadata'
    namespace = 'urn:xmpp:avatar:metadata'
    plugin_attrib = 'avatar_metadata'
    interfaces = set()

    def add_info(self, id, itype, ibytes, height=None, width=None, url=None):
        info = Info()
        info.values = {'id': id,
                       'type': itype,
                       'bytes': '%s' % ibytes,
                       'height': height,
                       'width': width,
                       'url': url}
        self.append(info)

    def add_pointer(self, xml):
        if not isinstance(xml, Pointer):
            pointer = Pointer()
            pointer.append(xml)
            self.append(pointer)
        else:
            self.append(xml)


class Info(ElementBase):
    name = 'info'
    namespace = 'urn:xmpp:avatar:metadata'
    plugin_attrib = 'info'
    plugin_multi_attrib = 'items'
    interfaces = {'bytes', 'height', 'id', 'type', 'url', 'width'}

    def _get_int(self, name: str) -> int:
        try:
            return int(self._get_attr(name))
        except ValueError:
            return 0

    def _set_int(self, name: str, value: int):
        if value not in ('', None):
            int(value)
        self._set_attr(name, value)

    def get_bytes(self) -> int:
        return self._get_int('bytes')

    def _set_bytes(self, value: int):
        self._set_int('bytes', value)

    def get_height(self) -> int:
        return self._get_int('height')

    def set_height(self, value: int):
        self._set_int('height', value)

    def get_width(self) -> int:
        return self._get_int('width')

    def set_width(self, value: int):
        self._set_int('width', value)


class Pointer(ElementBase):
    name = 'pointer'
    namespace = 'urn:xmpp:avatar:metadata'
    plugin_attrib = 'pointer'
    plugin_multi_attrib = 'pointers'
    interfaces = set()


register_stanza_plugin(MetaData, Info, iterable=True)
register_stanza_plugin(MetaData, Pointer, iterable=True)



# Slixmpp: The Slick XMPP Library
# Copyright (C) 2020 Mathieu Pasquet <mathieui@mathieui.net>
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
from slixmpp.plugins.base import register_plugin
from slixmpp.plugins.xep_0425.stanza import Moderate, Moderated
from slixmpp.plugins.xep_0425.moderation import XEP_0425

register_plugin(XEP_0425)

__all__ = ['Moderate', 'Moderated', 'XEP_0425']
